"""
  Support for Extensions, with support for Zeebe Properties

  <bpmn:collaboration id="...">
    <bpmn:extensionElements>
      <zeebe:properties>
        <zeebe:property name="name 1" value="value 1"></zeebe:property>
        <zeebe:property name="name 2" value="value 2"></zeebe:property>
      </zeebe:properties>
    </bpmn:extensionElements>
  </bpmn:collaboration>

  ExtensionElements
    +-- ExtensionGroup                --> ZeebeProperties
          +-- Extension
                +-- PropertyExtension --> ZeebePropertyExtension
"""

from . import xml

class ExtensionElements(xml.Element):
  __tag__ = "bpmn:extensionElements"

class ExtensionGroup(xml.Element):
  pass

class ZeebeProperties(ExtensionGroup):
  __tag__ = "zeebe:properties"

class Extension(xml.Element):
  pass

class PropertyExtension(Extension):
  def __init__(self, name=None, value=None):
    super().__init__()
    if name:
      self["name"] = name
    if value:
      self["value"] = value

class ZeebePropertyExtension(PropertyExtension):
  __tag__             = "zeebe:property"
  __extension_group__ = ZeebeProperties



"""
  Classes representing the different parts of a BPMN file.
"""

import logging

from .    import xml
from .xml import IdentifiedElement

logger = logging.getLogger(__name__)

class Flow(IdentifiedElement):
  __tag__ = "bpmn:sequenceFlow"

  def __init__(self, source=None, target=None, name=None, **kwargs):
    super().__init__(**kwargs)
    self._source = source
    self._target = target
    self["name"] = name

    if self._source:
      self._source.outgoing.append(Outgoing(self))
    if self._target:
      self._target.incoming.append(Incoming(self))
    self._resolving_source = False
    self._resolving_target = False

  @property
  def source(self):
    if self._source:
      return self._source
    elif self["sourceRef"]:
      if not self._resolving_source:
        self._resolving_source = True
        obj = self.root.find("id", self["sourceRef"])
        if obj:
          self._resolving_source = False
          return obj
      logger.warning(f"sourceRef {self['sourceRef']} not found.")
    else:
      logger.warning("no source available.")
    return None

  @property
  def target(self):
    if self._target:
      return self._target
    elif self["targetRef"]:
      if not self._resolving_target:
        self._resolving_target = True
        obj = self.root.find("id", self["targetRef"])
        if obj:
          self._resolving_target = False
          return obj
        logger.warning(f"targetRef {self['targetRef']} not found.")
    else:
      logger.warning("no target available.")
    return None

  def _get_flow_ids(self, default=None):
    try:
      source_id = self.source["id"]
    except Exception:
      source_id = default
    try:
      target_id = self.target["id"]
    except Exception:
      target_id = default
    return (source_id, target_id)

  def __getitem__(self, name):
    value = super().__getitem__(name)
    if value is None and name == "id":
      source_id, target_id = self._get_flow_ids(default="unknown")
      return f"flow_{source_id}_{target_id}"
    return value

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    attributes["id"] = self["id"]
    source_id, target_id = self._get_flow_ids()
    if source_id:
      attributes["sourceRef"] = source_id
    if target_id:
      attributes["targetRef"] = target_id
    if self.name:
      attributes["name"] = self.name
    return attributes

class MessageFlow(Flow):
  __tag__ = "bpmn:messageFlow"

class Incoming(xml.Element):
  __tag__ = "bpmn:incoming"
  
  def __init__(self, flow=None):
    super().__init__()
    self.flow = flow
    if self.flow:
      self.text = self.flow["id"]

class Outgoing(xml.Element):
  __tag__ = "bpmn:outgoing"

  def __init__(self, flow=None):
    super().__init__()
    self.flow = flow
    if self.flow:
      self.text = self.flow["id"]

  @property
  def target(self):
    if not self.flow:
      self.flow = self.root.find("id", self.text)
    return self.flow.target

class Element(IdentifiedElement):
  __labeled__ = True

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.incoming = []
    self.outgoing = []
    self.x      = 0
    self.y      = 0
    self.width  = 100
    self.height = 80

  @property
  def process(self):
    if isinstance(self._parent, Process):
      return self._parent
    if isinstance(self._parent, FlowNodeRef):
      # process -> (laneset) -> lane -> FlowNodeRef -> self
      return self._parent._parent._parent
    raise Exception("this should not happen ;-)")

  @property
  def lane(self):
    for lane in self.process.laneset.lanes:
      for ref in lane.refs:
        if ref.ref == self:
          return lane
    return None

  def append(self, child):
    if type(child) == Incoming:
      self.incoming.append(child)
    elif type(child) == Outgoing:
      self.outgoing.append(child)
    else:
      raise ValueError(f"Element {self} expects only incoming or outgoing flows. Got {type(child)}")
    child._parent = self
    return self

  @property
  def children(self):
    children = []
    if self.incoming:
      children.extend([ flow for flow in self.incoming ])
    if self.outgoing:
      children.extend([ flow for flow in self.outgoing ])
    return children

class EventDefinition(Element):
  pass

class MessageEventDefinition(EventDefinition):
  __tag__ = "bpmn:messageEventDefinition"

class TimerEventDefinition(EventDefinition):
  __tag__ = "bpmn:timerEventDefinition"

class EscalationEventDefinition(EventDefinition):
  __tag__ = "bpmn:escalationEventDefinition"

class ErrorEventDefinition(EventDefinition):
  __tag__ = "bpmn:errorEventDefinition"

class SignalEventDefinition(EventDefinition):
  __tag__ = "bpmn:signalEventDefinition"

class CompensateEventDefinition(EventDefinition):
  __tag__ = "bpmn:compensateEventDefinition"

class Condition(Element):
  __tag__ = "bpmn:condition"
  @property
  def attributes(self):
    return { "xsi:type" : "bpmn:tFormalExpression" }

class ConditionalEventDefinition(EventDefinition):
  """
  <bpmn:boundaryEvent id="Event_0lamv1c" attachedToRef="Activity_1sm8cz7">
    <bpmn:conditionalEventDefinition id="ConditionalEventDefinition_1224zru">
      <bpmn:condition xsi:type="bpmn:tFormalExpression" />
    </bpmn:conditionalEventDefinition>
  </bpmn:boundaryEvent>
  """
  __tag__ = "bpmn:conditionalEventDefinition"

  @property
  def children(self):
    return [ Condition() ]

  def append(self, child):
    if type(child) == Condition:
      return self
    return super().append(child)

EventDefinitions = [
  MessageEventDefinition,
  TimerEventDefinition,
  EscalationEventDefinition,
  ErrorEventDefinition,
  SignalEventDefinition,
  CompensateEventDefinition,
  ConditionalEventDefinition
]

class Event(Element):
  __labeled__ = True

  def __init__(self, name=None, definition=None, **kwargs):
    super().__init__(**kwargs)
    self.width   = 36
    self.height  = 36
    if name:
      self["name"] = name
    self.definition = definition

  def append(self, child):
    if isinstance(child, EventDefinition):
      self.definition = child
    else:
      super().append(child) # something else
    return self

  @property
  def children(self):
    children = super().children
    if self.definition:
      children.append(self.definition)
    return children

class Start(Event):
  __tag__    = "bpmn:startEvent"

class IntermediateThrow(Event):
  __tag__ = "bpmn:intermediateThrowEvent"

class IntermediateCatch(Event):
  __tag__ = "bpmn:intermediateCatchEvent"

class End(Event):
  __tag__ = "bpmn:endEvent"

# Boundary Events

class BoundaryEvent(Event):
  __tag__ = "bpmn:boundaryEvent"

  def __init__(self, on=None, **kwargs):
    super().__init__(**kwargs)
    self._on = on

  @property
  def on(self):
    if self._on:
      return self._on
    return self.root.find("id", self["attachedToRef"], skip=self, skip_classes=["Diagram"])

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.on:
      attributes["attachedToRef"] = self.on.id
    return attributes

  @property
  def x(self):
    if self.on:
      return self.on.x + self.on.width - int(self.width / 2)
    return 0

  @x.setter
  def x(self, value):
    pass

  @property
  def y(self):
    if self.on:
      return self.on.y - int(self.height / 2)
    return 0

  @y.setter
  def y(self, value):
    pass

# Tasks

class Task(Element):
  __tag__     = "bpmn:task"

  def __init__(self, name="", **kwargs):
    super().__init__(**kwargs)
    self["name"] = name

class UserTask(Task):
  __tag__ = "bpmn:userTask"

class ServiceTask(Task):
  __tag__ = "bpmn:serviceTask"

class SendTask(Task):
  __tag__ = "bpmn:sendTask"

class ReceiveTask(Task):
  __tag__ = "bpmn:receiveTask"

class ManualTask(Task):
  __tag__ = "bpmn:manualTask"

class BusinessRuleTask(Task):
  __tag__ = "bpmn:businessRuleTask"

class ScriptTask(Task):
  __tag__ = "bpmn:scriptTask"

Tasks = [
  Task,
  UserTask,
  ServiceTask,
  SendTask,
  ReceiveTask,
  ManualTask,
  BusinessRuleTask,
  ScriptTask
]

# Annotations

class Text(xml.Element):
  __tag__ = "bpmn:text"

class Annotation(IdentifiedElement):
  """
    <bpmn:textAnnotation id="TextAnnotation_1w5b2e0">
      <bpmn:text>comments</bpmn:text>
    </bpmn:textAnnotation>
  """
  
  __tag__ = "bpmn:textAnnotation"

  def __init__(self, text=None, **kwargs):
    super().__init__(**kwargs)
    self.annotation = text
  
  def append(self, child):
    if isinstance(child, Text):
      self.annotation = child.text
    else:
      super().append(child) # something else
    return self

  @property
  def children(self):
    return [ Text(text=self.annotation) ]

class Association(IdentifiedElement):
  __tag__ = "bpmn:association"

  def __init__(self, source=None, target=None, **kwargs):
    super().__init__(**kwargs)
    self._source = source
    self._target = target

    self._resolving_source = False
    self._resolving_target = False

  @property
  def source(self):
    if self._source:
      return self._source
    elif self["sourceRef"]:
      if not self._resolving_source:
        self._resolving_source = True
        obj = self.root.find("id", self["sourceRef"])
        if obj:
          self._resolving_source = False
          return obj
      logger.warning(f"sourceRef {self['sourceRef']} not found.")
    else:
      logger.warning("no source available.")
    return None

  @property
  def target(self):
    if self._target:
      return self._target
    elif self["targetRef"]:
      if not self._resolving_target:
        self._resolving_target = True
        obj = self.root.find("id", self["targetRef"])
        if obj:
          self._resolving_target = False
          return obj
        logger.warning(f"targetRef {self['targetRef']} not found.")
    else:
      logger.warning("no target available.")
    return None

  def _get_flow_ids(self, default=None):
    try:
      source_id = self.source["id"]
    except Exception:
      source_id = default
    try:
      target_id = self.target["id"]
    except Exception:
      target_id = default
    return (source_id, target_id)

  def __getitem__(self, name):
    value = super().__getitem__(name)
    if value is None and name == "id":
      source_id, target_id = self._get_flow_ids(default="unknown")
      return f"association_{source_id}_{target_id}"
    return value

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    attributes["id"] = self["id"]
    source_id, target_id = self._get_flow_ids()
    if source_id:
      attributes["sourceRef"] = source_id
    if target_id:
      attributes["targetRef"] = target_id
    return attributes

# Gateways

class Gateway(Element):
  __marker__  = True
  __labeled__ = True
  
  def __init__(self, name=None, **kwargs):
    super().__init__(**kwargs)
    self.width  = 50
    self.height = 50
    self["name"] = name

class ExclusiveGateway(Gateway):
  __tag__ = "bpmn:exclusiveGateway"

class InclusiveGateway(Gateway):
  __tag__ = "bpmn:inclusiveGateway"

class ParallelGateway(Gateway):
  __tag__ = "bpmn:parallelGateway"

Gateways = [
  ExclusiveGateway,
  InclusiveGateway,
  ParallelGateway
]

class FlowNodeRef(xml.Element):
  __tag__ = "bpmn:flowNodeRef"

  def __init__(self, ref=None, **kwargs):
    super().__init__(**kwargs)
    self._ref = ref

  @property
  def process(self):
    return self._parent.process

  @property
  def ref(self):
    if self._ref:
      return self._ref
    try:
      return self.process.element(super().text)
    except TypeError:
      pass
    return None
    
  @xml.Element.text.getter
  def text(self):
    # if we have referenced object (possibly resolved from #text)
    if self.ref:
      ref = self.ref.id
    else:
      ref = super().text    # else we should have a textual
    return ref

class Lane(IdentifiedElement):
  __tag__ = "bpmn:lane"
  __horizontal__ = True

  def __init__(self, name=None, *args, **kwargs):
    super().__init__(*args, **kwargs)
    if name:
      self["name"] = name
    self.refs    = []
    self.x      = 30
    self.y      = 0
    self.height = 125
    self.width  = 570

  @property
  def process(self):
    return self._parent.process

  @property
  def elements(self):
    # deref references
    return [ ref.ref for ref in self.refs if ref.ref ] # FIXME: ref.ref is None

  def append(self, child):
    if isinstance(child, FlowNodeRef):       # "native" ref
      self.refs.append(child)
      child._parent = self
    elif isinstance(child, Element):
      wrapped = FlowNodeRef(child)
      child._parent = wrapped
      wrapped._parent = self
      self.refs.append(wrapped)  # wrap it
    else:
      super().append(child) # something else
    return self

  @property
  def children(self):
    return super().children + self.refs

  def has_child(self, child):
    for ref in self.refs:
      if ref.ref == child:
        return True
    return False

class LaneSet(IdentifiedElement):
  __tag__ = "bpmn:laneSet"

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.lanes = []
  
  @property
  def process(self):
    return self._parent
  
  def append(self, child):
    if isinstance(child, Lane):
      self.lanes.append(child)
      child._parent = self
    else:
      super().append(child)
    return self
  
  def __len__(self):
    return len(self.lanes)

  @property
  def children(self):
    return super().children + self.lanes

  def lane_of(self, child):
    for lane in self.lanes:
      if lane.has_child(child):
        return lane
    return None

class Process(IdentifiedElement):
  __tag__ = "bpmn:process"
  
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.laneset = LaneSet(id=f"LaneSet_{self['id']}")
    self._elements = []
  
  def element(self, id):
    for element in self._elements:
      if element["id"] == id:
        return element
    return None
  
  @property
  def participant(self):
    for collaboration in self._parent.collaborations:
      participant = collaboration.find("processRef", self.id)
      if participant:
        return participant
    return None
  
  @property
  def elements(self):
    elems = self._elements  # direct elements
    # add elements that are "hidden" in lanes
    if len(self.laneset) > 0:
      for lane in self.laneset.lanes:
        for element in lane.elements:
          if element not in elems:
            elems.append(element)
    return elems
  
  def append(self, child):
    if isinstance(child, Element):
      self._elements.append(child)
      child._parent = self
    elif isinstance(child, Annotation):
      self._elements.append(child)
      child._parent = self
    elif isinstance(child, LaneSet):
      self.laneset = child
      child._parent = self
    elif isinstance(child, Lane):
      self.laneset.append(child)
      child._parent = self
    else:
      super().append(child) # generic child handling
    return self

  @property
  def children(self):
    children = super().children.copy()
    children.extend(self.elements)
    if len(self.laneset) > 0:
      children.append(self.laneset)
    return children



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = u'BPMN Tools'
copyright = u'2023-2023, Christophe VG'
author = u'Christophe VG'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser"] #, "autodoc2"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# autodoc2_render_plugin = "myst"
#
# autodoc2_packages = [
#   {
#     "path": "../bpmn_tools",
#     "exclude_dirs" : [ "templates" ],
#     "auto_mode": False
#   }
# ]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'

html_static_path = ["_static"]



from pathlib import Path
import xmltodict

from bpmn_tools.notation import Definitions

def test_process_of_event():
  """
    fixes issue that `process` property was at Task level
    by moving it to `Element` level, along with `lane` property
  """
  with open(Path(__file__).resolve().parent / "models" / "message-events.bpmn") as fp:
    xml = fp.read()
    d = xmltodict.parse(xml)
    definitions = Definitions.from_dict(d)
    end_message = definitions.find("name", "End Message")
    assert end_message.process == definitions.processes[0]



from pathlib import Path

from bpmn_tools.builder.process import Process, Task, Branch, BranchKind, If
from bpmn_tools.builder.process import connect

from bpmn_tools import flow
from bpmn_tools.colors import Green, Blue, Red, Orange

folder = Path(__file__).resolve().parent / "models"

def test_basic_process_building(compare_model_to_file):
  """
                        | -> [ Task 2 ] -------------- |
  S -> [ Task 1 ] -> < + >                            < > -> [ Task 4 ] -> S
                        | -> < x > ------------- < > - | 
                                | -> [ Task 3 ] - |
  """
  process = Process(name="main", starts=True, ends=True, color=Red).extend([
    Task(name="Task 1", color=Green),
    Branch(kind=BranchKind.AND, label="AND").add(
      Task(name="Task 2"), "with task 2"
    ).add(
      Branch(default=True, label="OR", color=Blue).add(
        Task(name="Task 3", color=Orange), "only task 3"
      ), "with task 3"
    ),
    Task(name="Task 4")
  ])

  model = process.render()
  
  filepath = folder / "hello-process-builder.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_complex_process_building(compare_model_to_file):
  process = Process([
    Task(name="Task 1"),
    Branch([
      Process([
        Branch([
          Task(name="Task 1b"),
          Task(name="Task 1c"),
          Task(name="Task 1d")
        ]),
        Task(name="Task 2a")
      ]),
      Task(name="Task 2b"),
      Process([
        Branch([
          Task(name="Task 2c-1"),
          Task(name="Task 2c-2"),
          Task(name="Task 2c-3")
        ]),
        Branch([
          Task(name="Task 2d-1"),
          Task(name="Task 2d-2"),
          Task(name="Task 2d-3")
        ])
      ]),
      Branch([
        Task(name="Task 3")
      ], default=True)
    ], kind=BranchKind.AND),
    Task(name="Task 4")
  ], name="complex", starts=True, ends=True)

  model = process.render()
  
  filepath = folder / "process-builder-complex-process-builder.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_bug_process_events_with_intial_branch(compare_model_to_file):
  process = Process([
    Branch([
      Task(name="Task 1"),
      Task(name="Task 2"),
      Task(name="Task 3")
    ]),
    Task(name="Task 4")
  ], name="bug start/end event with branch height", starts=True, ends=True)

  model = process.render()
  
  filepath = folder / "process-builder-bug_start_before_branch.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_branch_conditions_and_labels(compare_model_to_file):
  process = Process([
    Branch([
      If("long condition task 1", Task(name="Task 1")),
      If("medium task 2", Task(name="Task 2")),
      If("task 3", Task(name="Task 3"))
    ], default=True, label="conditions"),
    Task(name="Task 4")
  ], name="conditions and labels", starts=True, ends=True)

  model = process.render()
  
  filepath = folder / "process-builder-conditions_and_labels.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_boundary_event(compare_model_to_file):
  process = Process([
    Task(name="Task 1", boundary=flow.MessageEventDefinition)
  ], name="task with boundary event")

  model = process.render()
  
  filepath = folder / "process-builder-based-boundary-events.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_bug_flow_from_boundary_event_to_end(compare_model_to_file):
  process = Process([
    Branch([
      Task(name="Task 1", boundary=flow.MessageEventDefinition)
    ], default=True)
  ], name="task with boundary event", ends=True)

  model = process.render()
  
  filepath = folder / "process-builder-bug_flow_from_boundary_event_to_end.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_task_id():
  task = Task()                             # task_0
  assert task.id == "task_0"
  assert task.element.id == "task_0"

  task = Task(id="test_task")               # task_1
  assert task.id == "test_task"
  assert task.element.id == "test_task"

  task = Task(id=lambda d: f"prefixed_{d}") # task_2
  assert task.id == "prefixed_task_2"
  assert task.element.id == "prefixed_task_2"

def test_connect_only_accepts_different_source_and_target_shapes():
  task = flow.Task(name="task")
  try:
    connect(task, task)
    assert False, "connect should not accept same source and target"
  except ValueError:
    pass



import pytest

from pathlib import Path
from difflib import unified_diff
import json

from bpmn_tools.util import model2xml, sanitize_xml

from bpmn_tools.builder.process import Task, Branch

folder = Path(__file__).resolve().parent / "models"

def compare_strings(result, expected):
  # print to std, which shows up in test output when assertion fails
  print("".join(unified_diff(
    expected.splitlines(keepends=True),
    result.splitlines(keepends=True)
  )))
  assert result == expected, "see diff"

def compare(result,expected):
  result_str   = json.dumps(result,   indent=2, sort_keys=True)
  expected_str = json.dumps(expected, indent=2, sort_keys=True)
  compare_strings(result_str, expected_str)

def load_model_src(filename):
  return (folder / filename).read_text()

@pytest.fixture(name="model")
def model_fixture():
  def fixture(filename):
    return load_model_src(filename)
  return fixture

def compare_to_file(result, filename, save_to=None):
  if save_to:
    (folder / save_to).write_text(result)
  try:
    expected = load_model_src(filename)
    expected = sanitize_xml(expected)
  except FileNotFoundError:
    expected = ""
  compare_strings(result, expected)

def compare_model_to_file(model, filename, save_to=None):
  xml = model2xml(model)
  compare_to_file(xml, filename, save_to)
  return xml

@pytest.fixture(name="compare_to_file")
def compare_to_file_fixture():
  def tester(result, filename, save_to=None) :
    return compare_to_file(result, filename, save_to=save_to)
  return tester

@pytest.fixture(name="compare")
def compare_fixture():
  def tester(result, expected):
    return compare(result, expected)
  return tester

@pytest.fixture(name="compare_model_to_file")
def compare_model_to_file_fixture():
  def tester(model, filename, save_to=None):
    return compare_model_to_file(model, filename, save_to=save_to)
  return tester

@pytest.fixture(autouse=True)
def ensure_starting_each_test_with_generated_ids_starting_at_0():
  Task.reset()
  Branch.reset()
  yield



from pathlib import Path
import xmltodict

from bpmn_tools.xml     import Element
from bpmn_tools.visitor import Visitor

def test_visiting_hello():
  with open(Path(__file__).resolve().parent / "models" / "hello.bpmn") as fp:
    xml = xmltodict.parse(fp.read())
    tree = Element.from_dict(xml)

    visitations = []

    class MyVisitor(Visitor):
      def visit(self, visited):
        if visited["id"]:
          visitations.append(visited["id"])

    tree.accept(MyVisitor())

    assert visitations == [
      'definitions',
      'process',
      'flow_start_hello',
      'flow_hello_end',
      'start',
      'hello',
      'end',
      'collaboration',
      'participant',
      'diagram',
      'plane_collaboration',
      'edge_flow_hello_end',
      'edge_flow_start_hello',
      'shape_end',
      'shape_hello',
      'shape_participant',
      'shape_start'
    ]



import logging
import os
import sys

from fire import Fire

from bpmn_tools.tool import CLI

logger = logging.getLogger(__name__)

LOG_LEVEL = os.environ.get("LOG_LEVEL") or "DEBUG"
FORMAT    = "[%(asctime)s] [%(name)s] [%(process)d] [%(levelname)s] %(message)s"
DATEFMT   = "%Y-%m-%d %H:%M:%S %z"
logging.basicConfig(level=LOG_LEVEL, format=FORMAT, datefmt=DATEFMT)
formatter = logging.Formatter(FORMAT, DATEFMT)

def cli():
  try:
    Fire(CLI(), name="bpmn-tool", command=sys.argv[1:] + ["--", "--separator=XXX"])
  except KeyboardInterrupt:
    pass

if __name__ == "__main__":
  cli()



import pytest

from pathlib import Path
import xmltodict

from bpmn_tools.notation  import Definitions

from tests.conftest import compare_model_to_file

def perform_roundtrip_test(filepath):
  """
  given a filename, the BPMN XML is 
  1. loaded as a dict
  2. used to construct a Definitions object
  3. of which the XML rendering is compared to the original
  """
  xml = filepath.read_text()

  # xml -> obj
  as_dict = xmltodict.parse(xml)
  model   = Definitions.from_dict(as_dict, raise_unmapped=True)
  
  # ensure that rebuild model from parsed xml is still equal
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-roundtrip.bpmn")

folder = Path(__file__).resolve().parent / "models" 
models = [ str(filepath.name) for filepath in folder.glob("*.bpmn") ]

@pytest.mark.parametrize("filename", models)
def test_all_reference_models(filename):
  filepath = folder / filename
  if not filepath.stem.endswith("-roundtrip") and not filepath.stem.endswith("-test") and "goal" not in filepath.stem:
    if filepath.stem.startswith("bad-"):
      try:
        perform_roundtrip_test(filepath)
        assert False, f"expected exception: {filepath.stem}"
      except Exception:
        pass
    else:
      perform_roundtrip_test(filepath)



from bpmn_tools.layout import graphviz
from bpmn_tools import util

from tests.conftest import load_model_src
from unittest.mock import Mock
import json

def test_graphviz_routing(compare_model_to_file):
  model = util.xml2model(load_model_src("graphviz-original.bpmn"))

  # graphviz doesn't produce the same results on different platforms, so, to
  # eliminate this, we're mocking this
  gv_mock = Mock()
  gv_mock.pipe.return_value = json.dumps({
    "objects": [
      { "name": "Activity_0yvzjff", "pos": "69.544,380.58" },
      { "name": "Activity_0k2y43l", "pos": "182.25,60.475" },
      { "name": "Activity_00ip78i", "pos": "419.35,603.23" },
      { "name": "Activity_12mr32b", "pos": "394.39,300.6"  },
      { "name": "Activity_1i04bed", "pos": "526.28,18"     },
      { "name": "Activity_1t7zz52", "pos": "715.2,319.19"  }
    ]
  }).encode("ascii")

  graphviz.layout(model, g=gv_mock)
  compare_model_to_file(model, "graphviz-goal.bpmn", save_to="graphviz-goal-test.bpmn")



from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Start, End, UserTask, ScriptTask
from bpmn_tools.flow          import Flow, Lane
from bpmn_tools.flow          import MessageFlow
from bpmn_tools.diagrams      import Diagram, Plane

from bpmn_tools.layout        import simple

def test_diagram_with_lane(compare_model_to_file):
  activities_lane1 = [
    Start(id="start"),
    UserTask('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  activities_lane2 = [
    ScriptTask('Hear "Hello!"', id="hear"),
    ScriptTask('Also Hear "Hello!"', id="hear2"),
    ScriptTask('Also Also Hear "Hello!"', id="hear3"),
  ]

  lane2 = Lane("lane 2", id="lane2").extend(activities_lane2)

  process_lane1 = Process(id="process1").extend(activities_lane1).extend([
    Flow(id="flow_start_hello", source=activities_lane1[0], target=activities_lane1[1]),
    Flow(id="flow_hello_end",   source=activities_lane1[1], target=activities_lane1[2])
  ])

  process_lane2 = Process(id="process2").append(lane2)

  collaboration = Collaboration(id="collaboration").extend([
    Participant("participant1", process_lane1, id="participant1"),
    Participant("participant2", process_lane2, id="participant2"),
    MessageFlow(id="flow_hello_hear",  source=activities_lane1[1], target=activities_lane2[0]),
    MessageFlow(id="flow_hello_hear2", source=activities_lane1[1], target=activities_lane2[1]),
    MessageFlow(id="flow_hello_hear3", source=activities_lane1[1], target=activities_lane2[2])
  ])

  definitions = Definitions(id="definitions").extend([
    process_lane1,
    process_lane2,
    collaboration
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  simple.layout(definitions)

  compare_model_to_file(
    definitions,
    "hello-lanes-with-message.bpmn",
    save_to="hello-lanes-with-message-test.bpmn"
  )
  
  # TODO: first implement extraction of x, y, height, width from Shape->Element
  # compare_with_roundtrip(definitions, expected)



import xmltodict

from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Task
from bpmn_tools.diagrams      import Diagram, Plane 

from bpmn_tools.flow          import BoundaryEvent
from bpmn_tools.flow          import EventDefinitions

def test_message_boundaries(compare_model_to_file):
  X = 130
  Y =  60

  activities = []
  for index, event_definition in enumerate(EventDefinitions):
    name = event_definition.__name__
    task = Task(name, id=f"task-{name}")
    task.x = X + 60 + (task.width + 20) * index
    task.y = Y + 30
    event = BoundaryEvent(
      id=f"boundary-event-{name}",
      definition=event_definition(id=f"message-event-definition-{name}"),
      on=task
    )
    activities.extend([task, event])

  process = Process(id="process").extend(activities)

  participant = Participant("lane", process, id="participant")
  participant.x = X
  participant.y = Y
  participant.width = len(EventDefinitions) * (100 + 20) + 70 # TODO fix ;-)

  collaboration = Collaboration(id="collaboration").append(participant)

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  compare_model_to_file(definitions, "boundaries.bpmn", save_to="boundaries-test.bpmn")

def test_dangling_boundaries(model):
  xml = model("bad-boundaries.bpmn")
  d = xmltodict.parse(xml)
  definitions = Definitions.from_dict(d)
  for event in definitions.children_oftype(BoundaryEvent, recurse=True):
    try:
      assert event.on
      assert event.id != "bad-boundary-event", "dangling boundary should have None"
    except AssertionError:
      pass



from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Task, Flow
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.colors        import Red, Orange, Green, Blue
from bpmn_tools.layout        import simple

def test_colors(compare_model_to_file):
  activities = [
    Red(Task('Red"',    id="red")),
    Green(Task('Green', id="green")),
    Blue(Task('Blue"',  id="blue"))
  ]

  process = Process(id="process").extend(activities).extend([
    Flow(id="flow_red_green",  source=activities[0], target=activities[1]),
    Flow(id="flow_green_blue", source=activities[1], target=activities[2])
  ])

  collaboration = Collaboration(id="collaboration").append(
    Orange(Participant("lane", process, id="participant"))
  )

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )
  
  simple.layout(definitions)

  compare_model_to_file(definitions, "hello-colors.bpmn", save_to="hello-colors-test.bpmn")



from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, IntermediateThrow, MessageEventDefinition
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.layout        import simple

def test_message_event(compare_model_to_file):
  message = MessageEventDefinition(id="message")
  event = IntermediateThrow(id="event", definition=message)

  process = Process(id="process").append(event)

  collaboration = Collaboration(id="collaboration").append(
    Participant("lane", process, id="participant")
  )

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )
  
  simple.layout(definitions)

  compare_model_to_file(definitions, "events.bpmn", save_to="events-test.bpmn")



"""
  Classes representing the different parts of a BPMN file.
"""

import sys
import inspect

from bpmn_tools.collaboration import Collaboration
from bpmn_tools.flow          import Process
from bpmn_tools.diagrams      import Diagram

from bpmn_tools.xml import Element, IdentifiedElement

from bpmn_tools.layout import routing

def get_classes(module):
  return [
    c[1] for c in inspect.getmembers(sys.modules[module], inspect.isclass)
  ]

class Definitions(IdentifiedElement):
  __tag__ = "bpmn:definitions"

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.processes        = []
    self.collaborations   = []
    self.diagrams         = []
    self.routing_strategy = routing.Default()

  @classmethod
  def from_dict(cls, d, raise_unmapped=False):
    classes = get_classes("bpmn_tools.notation") + \
              get_classes("bpmn_tools.collaboration") + \
              get_classes("bpmn_tools.flow") + \
              get_classes("bpmn_tools.diagrams") + \
              get_classes("bpmn_tools.extensions") + \
              [ routing.WayPoint ]
    definitions = Element.from_dict(d, classes=classes, raise_unmapped=raise_unmapped)
    if isinstance(definitions, Definitions):
      for diagram in definitions.diagrams:
        for shape in diagram.plane._shapes:
          if shape._bounds:
            element = definitions.find("id", shape["bpmnElement"])
            if element:
              element.x      = int(float(shape._bounds.x))
              element.y      = int(float(shape._bounds.y))
              element.width  = int(float(shape._bounds.width))
              element.height = int(float(shape._bounds.height))
              color_scheme   = {}
              for k, v in shape._attributes.items():
                if k in [
                  "bioc:stroke",
                  "bioc:fill",
                  "color:background-color",
                  "color:border-color"
                ]:
                  color_scheme[k] = v
              if color_scheme:
                element.__color_scheme__ = color_scheme
            else:
              raise ValueError(f"diagram shape ({shape}) has no element")
    return definitions

  def append(self, child):
    if isinstance(child, Process):
      self.processes.append(child)
      child._parent = self
    elif isinstance(child, Collaboration):
      self.collaborations.append(child)
      child._parent = self
    elif isinstance(child, Diagram):
      self.diagrams.append(child)
      child._parent = self
    else:
      raise ValueError(f"unsupported child for Definitions: {child}")
    return self
    return self

  def element(self, id):
    for process in self.processes:
      match = process.find("id", id)
      if match:
        return match
    return None
    
  @property
  def children(self):
    return super().children + self.processes + self.collaborations + self.diagrams

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    attributes.update({
      "xmlns:bpmn"  : "http://www.omg.org/spec/BPMN/20100524/MODEL",
      "xmlns:bpmndi": "http://www.omg.org/spec/BPMN/20100524/DI",
      "xmlns:dc"    : "http://www.omg.org/spec/DD/20100524/DC",
      "xmlns:di"    : "http://www.omg.org/spec/DD/20100524/DI",
      "xmlns:bioc"  : "http://bpmn.io/schema/bpmn/biocolor/1.0",
      "xmlns:color" : "http://www.omg.org/spec/BPMN/non-normative/color/1.0",
      "xmlns:zeebe" : "http://camunda.org/schema/zeebe/1.0",
      "xmlns:xsi"   : "http://www.w3.org/2001/XMLSchema-instance"
    })
    return attributes

  def apply(self, strategy):
    for diagram in self.diagrams:
      diagram.strategy = strategy



"""
  Classes representing the different parts of a BPMN file.
"""

from . import xml

from bpmn_tools.collaboration import Participant
from bpmn_tools.flow          import Element, Flow, MessageFlow
from bpmn_tools.flow          import Annotation, Association

from bpmn_tools.layout         import routing

class Bounds(xml.Element):
  __tag__ = "dc:Bounds"
  
  def __init__(self, x=0, y=0, width=0, height=0):
    super().__init__()
    self["x"]      = str(int(x))
    self["y"]      = str(int(y))
    self["width"]  = str(int(width))
    self["height"] = str(int(height))

class Label(xml.Element):
  __tag__ = "bpmndi:BPMNLabel"
  
  def __init__(self, x=0, y=0):
    super().__init__()
    self.x = x - 50         # add offset from target point x,y
    self.y = y + 5

  @property
  def children(self):
    return [
      Bounds(
        x=self.x,
        y=self.y,
        width=0,
        height=14
      )
    ]

class Shape(xml.Element):
  __tag__ = "bpmndi:BPMNShape"

  def __init__(self, element=None, id=None):
    super().__init__()
    self._element = element
    self._bounds  = None

  def append(self, child):
    if isinstance(child, Bounds):
      self._bounds = child
      child._parent = self
    else:
      super().append(child)
    return self

  @property
  def id(self):
    return f"shape_{self.element['id']}"

  @property
  def element(self):
    if self._element:
      return self._element
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  @property
  def attributes(self):
    attributes = {}
    if self.element:
      attributes.update({
        "id"          : self.id,
        "bpmnElement" : self.element["id"]
      })
      if self.element.__horizontal__:
        attributes["isHorizontal"] = "true"
      if self.element.__marker__:
        attributes["isMarkerVisible"] = "true"
      if self.element.__color_scheme__:
        attributes.update(self.element.__color_scheme__)
    return attributes

  @property
  def children(self):
    children = []
    if self.element:
      children = [
        Bounds(
          x      = self.element.x,
          y      = self.element.y,
          height = self.element.height,
          width  = self.element.width
        )
      ]
    return children

class Edge(xml.Element):
  __tag__ = "bpmndi:BPMNEdge"

  def __init__(self, flow=None, id=None, strategy=None):
    super().__init__()
    self._flow = flow
    self.strategy = routing.Default() if strategy is None else strategy

  @property
  def flow(self):
    if self._flow:
      return self._flow
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  def __getitem__(self, name):
    if name == "id":
      return f"edge_{self.flow['id']}"
    return super().__getitem__(name)

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.flow:
      attributes.update({
        "id"          : self["id"],
        "bpmnElement" : self.flow["id"]
      })
    return attributes

  @property
  def children(self):
    children = super().children.copy()

    if not self.flow or not self.flow.source or not self.flow.target:
      return children

    children = self.strategy.route(self.flow)

    if self.flow["name"]:
      x = self.flow.target.x
      y = self.flow.target.y + self.flow.target.height/2
      children.append(Label(x, y))
    return children

class Plane(xml.Element):
  __tag__ = "bpmndi:BPMNPlane"

  def __init__(self, id="plane", element=None, strategy=None):
    super().__init__()
    self._element = element
    self["id"]    = id
    self._shapes  = []
    self._edges   = []
    self.strategy = strategy

  @property
  def element(self):
    if self._element:
      return self._element
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.element:
      attributes.update({
        "bpmnElement" : self.element["id"],
        "id" : f"plane_{self.element['id']}"
      })
    return attributes

  def append(self, child):
    if isinstance(child, Shape):
      self._shapes.append(child)
      child._parent = self
    elif isinstance(child, Edge):
      self._edges.append(child)
      child._parent = self
    else:
      super().append(child)
    return self

  @property
  def children(self):
    children = super().children.copy()
    if self.element:
      for participant in self.element.children_oftype(Participant):
        children.append(Shape(participant))
        if participant.process:
          for lane in participant.process.laneset.lanes:
            children.append(Shape(lane))
          for element in participant.process.children_oftype(Element):
            children.append(Shape(element))
          for element in participant.process.children_oftype(Annotation):
            children.append(Shape(element))
          for element in participant.process.children_oftype(Association):
            children.append(self.edge(element))
          for flow in participant.process.children_oftype(Flow):
            children.append(self.edge(flow))
      for flow in self.element.children_oftype(MessageFlow):
        children.append(self.edge(flow))
    return sorted(children, key=lambda k: k.id)

  def edge(self, flow):
    return Edge(flow, strategy=self.strategy)
    
class Diagram(xml.Element):
  __tag__ = "bpmndi:BPMNDiagram"

  def __init__(self, id="diagram", plane=None):
    super().__init__()
    self._plane   = plane
    self["id"]    = id
    self.strategy = routing.Default()

  @property
  def plane(self):
    if self._plane:
      plane = self._plane
    elif self._children:
      plane = self._children[0]
    else:
      plane = Plane()
    plane.strategy = self.strategy
    return plane

  @property
  def children(self):
    return [ self.plane ]



"""
  Classes representing the different parts of a BPMN file.
"""

from . import xml

from .extensions import Extension, ExtensionElements, PropertyExtension

class Participant(xml.Element):
  __tag__        = "bpmn:participant"
  __horizontal__ = True

  def __init__(self, name="participant", process=None, id="participant"):
    super().__init__()
    self["id"]   = id
    self["name"] = name
    self._process = process
    self.x      = 0
    self.y      = 0
    self.height = 125
    self.width  = 600
  
  @property
  def process(self):
    if self._process:
      return self._process
    elif self["processRef"]:
      return self.root.find("id", self["processRef"])
    return None
  
  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.process:
      attributes.update({
        "processRef" : self.process["id"]
      })
    return attributes

class Collaboration(xml.Element):
  __tag__ = "bpmn:collaboration"

  def __init__(self, id="collaboration"):
    super().__init__()
    self["id"] = id
  
  @property
  def extension_properties(self):
    return {
      extension.name : extension.value
      for extension in self.children_oftype(cls=PropertyExtension, recurse=True)
    }
  
  def append(self, child):
    # accept short-hand appending of Extensions, wrapping them in
    # ExtensionElements > [...]Properties
    if isinstance(child, Extension):
      try:
        extension_elements = self.children_oftype(ExtensionElements)[0]
      except Exception:
        extension_elements = ExtensionElements()
        super().append(extension_elements)
      try:
        group = extension_elements.children_oftype(child.__extension_group__)[0]
      except Exception:
        group = child.__extension_group__()
        extension_elements.append(group)
    
      group.append(child)
    else:
      super().append(child)
    return self
  
  @property
  def children(self):
    # put ExtensionElements first in line ;-)
    children = super().children.copy()
    children.sort(key=lambda c: 0 if c.__class__ is ExtensionElements else 1)
    return children



from bpmn_tools.flow          import Process, Start, End, Task
from bpmn_tools.flow          import Flow

def test_create_single_step_process():
  task = Task('Say "Hello!"', id="hello")

  activities = [ Start(id="start"), task, End(id="end") ]

  process = Process(id="process").extend(activities).extend([
    Flow(source=activities[0], target=activities[1]),
    Flow(source=activities[1], target=activities[2])
  ])

  assert task.process == process



"""
  Initial tests to drive development entirely from tests ;-)
  TDD rules 💪
  The tests build towards a complete support to construct the hello.bpmn example
"""

from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Start, End, Task, Flow
from bpmn_tools.diagrams      import Diagram, Plane, Shape, Edge

def test_create_single_step_process(compare):
  """
    Create a single-step process.
    (start) -> (task: Say "Hello!") -> (end)
  
    <bpmn:process id="process" isExecutable="true">
      <bpmn:startEvent id="start">
        <bpmn:outgoing>flow_start_hello</bpmn:outgoing>
      </bpmn:startEvent>
      <bpmn:task id="hello" name="Say &#34;Hello!&#34;">
        <bpmn:incoming>flow_start_hello</bpmn:incoming>
        <bpmn:outgoing>flow_hello_end</bpmn:outgoing>
      </bpmn:task>
      <bpmn:endEvent id="end">
        <bpmn:incoming>flow_hello_end</bpmn:incoming>
      </bpmn:endEvent>
      <bpmn:sequenceFlow id="flow_start_hello" sourceRef="start" targetRef="hello" />
      <bpmn:sequenceFlow id="flow_hello_end" sourceRef="hello" targetRef="end" />
    </bpmn:process>
  """
  activities = [
    Start(id="start"),
    Task('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  process = Process(id="process").extend(activities).extend([
    Flow(id="flow_start_hello", source=activities[0], target=activities[1]),
    Flow(id="flow_hello_end",   source=activities[1], target=activities[2])
  ])

  compare(process.as_dict(with_tag=True), {
    "bpmn:process": {
      "@id": "process",
      "bpmn:startEvent": {
        "@id": "start",
        "bpmn:outgoing": "flow_start_hello"
      },
      "bpmn:task": {
        "@id": "hello",
        "@name": "Say \"Hello!\"",
        "bpmn:incoming": "flow_start_hello",
        "bpmn:outgoing": "flow_hello_end"
      },
      "bpmn:endEvent": {
        "@id": "end",
        "bpmn:incoming": "flow_hello_end"
      },
      "bpmn:sequenceFlow": [
        {
          "@id": "flow_start_hello",
          "@sourceRef": "start",
          "@targetRef": "hello"
        },
        {
          "@id": "flow_hello_end",
          "@sourceRef": "hello",
          "@targetRef": "end"
        }
      ]
    }
  })

def test_create_single_participant_collaboration(compare):
  """
    Create a single participant collaboration for Process(id="process")

    <bpmn:collaboration id="collaboration">
      <bpmn:participant id="participant" name="lane" processRef="process" />
    </bpmn:collaboration>
  """
  collaboration = Collaboration(id="collaboration").append(
    Participant("participant", Process(id="process"), id="participant")
  )

  compare(collaboration.as_dict(with_tag=True), {
    "bpmn:collaboration": {
      "@id": "collaboration",
      "bpmn:participant": {
        "@id": "participant",
        "@name": "participant",
        "@processRef": "process"
      }
    }
  })

def test_create_definitions_with_process_and_collaboration(compare):
  """
    Create Definitions with both a single-step process and collaboration.

    <bpmn:definitions xmlns:bpmn="http://www.omg.org/spec/BPMN/20100524/MODEL"
                      xmlns:bpmndi="http://www.omg.org/spec/BPMN/20100524/DI"
                      xmlns:dc="http://www.omg.org/spec/DD/20100524/DC"
                      xmlns:di="http://www.omg.org/spec/DD/20100524/DI"
                      xmlns:modeler="http://camunda.org/schema/modeler/1.0"
                      id="definitions"
                      xmlns:zeebe="http://camunda.org/schema/zeebe/1.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
      <bpmn:collaboration id="collaboration">
        <bpmn:participant id="participant" name="lane" processRef="process" />
      </bpmn:collaboration>
      <bpmn:process id="process" isExecutable="true">
        <bpmn:startEvent id="start">
          <bpmn:outgoing>flow_start_hello</bpmn:outgoing>
        </bpmn:startEvent>
        <bpmn:task id="hello" name="Say &#34;Hello!&#34;">
          <bpmn:incoming>flow_start_hello</bpmn:incoming>
          <bpmn:outgoing>flow_hello_end</bpmn:outgoing>
        </bpmn:task>
        <bpmn:endEvent id="end">
          <bpmn:incoming>flow_hello_end</bpmn:incoming>
        </bpmn:endEvent>
        <bpmn:sequenceFlow id="flow_start_hello" sourceRef="start" targetRef="hello" />
        <bpmn:sequenceFlow id="flow_hello_end" sourceRef="hello" targetRef="end" />
      </bpmn:process>
    </bpmn:definitions>
  """
  activities = [
    Start(id="start"),
    Task('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  process = Process(id="process").extend(activities).extend([
    Flow(id="flow_start_hello", source=activities[0], target=activities[1]),
    Flow(id="flow_hello_end",   source=activities[1], target=activities[2])
  ])

  collaboration = Collaboration(id="collaboration").append(
    Participant("participant", process, id="participant")
  )

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration
  ])

  compare(definitions.as_dict(with_tag=True), {
    "bpmn:definitions": {
      "@xmlns:bioc": "http://bpmn.io/schema/bpmn/biocolor/1.0",
      "@xmlns:color": "http://www.omg.org/spec/BPMN/non-normative/color/1.0",
      "@xmlns:bpmn": "http://www.omg.org/spec/BPMN/20100524/MODEL",
      "@xmlns:bpmndi": "http://www.omg.org/spec/BPMN/20100524/DI",
      "@xmlns:dc": "http://www.omg.org/spec/DD/20100524/DC",
      "@xmlns:di": "http://www.omg.org/spec/DD/20100524/DI",
      "@xmlns:zeebe": "http://camunda.org/schema/zeebe/1.0",
      "@xmlns:xsi" : "http://www.w3.org/2001/XMLSchema-instance",
      "@id": "definitions",
      "bpmn:process": {
        "@id": "process",
        "bpmn:startEvent": {
          "@id": "start",
          "bpmn:outgoing": "flow_start_hello"
        },
        "bpmn:task": {
          "@id": "hello",
          "@name": "Say \"Hello!\"",
          "bpmn:incoming": "flow_start_hello",
          "bpmn:outgoing": "flow_hello_end"
        },
        "bpmn:endEvent": {
          "@id": "end",
          "bpmn:incoming": "flow_hello_end"
        },
        "bpmn:sequenceFlow": [
          {
            "@id": "flow_start_hello",
            "@sourceRef": "start",
            "@targetRef": "hello"
          },
          {
            "@id": "flow_hello_end",
            "@sourceRef": "hello",
            "@targetRef": "end"
          }
        ]
      },
      "bpmn:collaboration": {
        "@id": "collaboration",
        "bpmn:participant": {
          "@id": "participant",
          "@name": "participant",
          "@processRef": "process"
        }
      }
    }
  })

def test_create_start_shape(compare):
  """
    Create a shape for a Start element, with default bounds

    <bpmndi:BPMNShape id="shape_start" bpmnElement="start">
      <dc:Bounds x="179" y="99" width="36" height="36" />
    </bpmndi:BPMNShape>
  """

  shape = Shape(Start(id="start"))

  compare(shape.as_dict(with_tag=True), {
    "bpmndi:BPMNShape": {
      "@id": "shape_start",
      "@bpmnElement": "start",
      "dc:Bounds": {
        "@x": "0",
        "@y": "0",
        "@width": "36",
        "@height": "36"
      }
    }
  })

def test_create_task_shape(compare):
  """
    Create a shape for a Task element, with default bounds

    <bpmndi:BPMNShape id="shape_hello" bpmnElement="hello">
      <dc:Bounds x="270" y="77" width="100" height="80" />
    </bpmndi:BPMNShape>
  """

  shape = Shape(Task(id="hello"))

  compare(shape.as_dict(with_tag=True), {
    "bpmndi:BPMNShape": {
      "@id": "shape_hello",
      "@bpmnElement": "hello",
      "dc:Bounds": {
        "@x": "0",
        "@y": "0",
        "@width": "100",
        "@height": "80"
      }
    }
  })

def test_create_edge(compare):
  """
    Create an edge for a flow between start and

    <bpmndi:BPMNEdge id="edge_flow_start_hello" bpmnElement="flow_start_hello">
      <di:waypoint x="215" y="117" />
      <di:waypoint x="270" y="117" />
    </bpmndi:BPMNEdge>
  """

  flow = Flow(id="flow_start_hello", source=Start(id="start"), target=Task(id="hello"))
  edge = Edge(flow)

  compare(edge.as_dict(with_tag=True), {
    "bpmndi:BPMNEdge": {
      "@id": "edge_flow_start_hello",
      "@bpmnElement": "flow_start_hello",
      "di:waypoint": [
        {
          "@x": "36",
          "@y": "18"
        },
        {
          "@x": "0",
          "@y": "40"
        }
      ]
    }
  })

def test_empty_diagram(compare):
  """
    An empty diagram, always contains a plane.

    <bpmndi:BPMNDiagram id="diagram">
      <bpmndi:BPMNPlane id="plane">
      </bpmndi:BPMNPlane>
    </bpmndi:BPMNDiagram>
  """

  diagram = Diagram(id="diagram")

  compare(diagram.as_dict(with_tag=True), {
    "bpmndi:BPMNDiagram": {
      "@id": "diagram",
      "bpmndi:BPMNPlane": {
        "@id": "plane"
      }
    }
  })

def test_plane_for_collaboration_with_one_participant_without_a_process(compare):
  """
    A plane for a collaboration with on participant.

    <bpmndi:BPMNPlane id="plane_collaboration" bpmnElement="collaboration">
      <bpmndi:BPMNShape id="participant_di" bpmnElement="participant" isHorizontal="true">
        <dc:Bounds x="129" y="57" width="600" height="123" />
      </bpmndi:BPMNShape>
    </bpmndi:BPMNPlane>
  """ # noqa
  collaboration = Collaboration(id="collaboration").append(
    Participant("participant", None, id="participant")
  )
  plane = Plane(id="plane", element=collaboration)

  compare(plane.as_dict(with_tag=True), {
    "bpmndi:BPMNPlane": {
      "@id": "plane_collaboration",
      "@bpmnElement": "collaboration",
      "bpmndi:BPMNShape": {
        "@id": "shape_participant",
        "@bpmnElement": "participant",
        "@isHorizontal": "true",
        "dc:Bounds": {
          "@x": "0",
          "@y": "0",
          "@width": "600",
          "@height": "125"
        }
      }
    }
  })

def test_hello(compare_model_to_file):
  activities = [
    Start(id="start"),
    Task('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  process = Process(id="process").extend(activities).extend([
    Flow(id="flow_start_hello", source=activities[0], target=activities[1]),
    Flow(id="flow_hello_end",   source=activities[1], target=activities[2])
  ])

  collaboration = Collaboration(id="collaboration").append(
    Participant("lane", process, id="participant")
  )

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  compare_model_to_file(definitions, "hello.bpmn", save_to="hello-test.bpmn")



import xmltodict

from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.extensions    import ZeebePropertyExtension

from tests.conftest import load_model_src

def test_minimal_zeebe_properies_extension(compare_model_to_file):

  process = Process(id="process")

  collaboration = Collaboration(id="collaboration").append(
    Participant("lane", process, id="participant")
  ).extend([
    ZeebePropertyExtension("name 1", "value 1"),
    ZeebePropertyExtension("name 2", "value 2")
  ])

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  compare_model_to_file(
    definitions, "extensions-zeebe.bpmn", save_to="extensions-zeebe-test.bpmn"
  )

def test_extensions_access():
  process = Process(id="process")

  collaboration = Collaboration(id="collaboration").append(
    Participant("lane", process, id="participant")
  ).extend([
    ZeebePropertyExtension("zeebe name 1", "value 1"),
    ZeebePropertyExtension("zeebe name 2", "value 2")
  ])
  
  assert collaboration.extension_properties == {
    "zeebe name 1"   : "value 1",
    "zeebe name 2"   : "value 2"
  }

def test_extensions_loading_from_xml():
  definitions = Definitions.from_dict(
    xmltodict.parse(load_model_src("extensions-zeebe.bpmn"))
  ) 

  assert definitions.collaborations[0].extension_properties == {
    "name 1" : "value 1",
    "name 2" : "value 2"
  }



import xmltodict
import re

from bpmn_tools.notation import Definitions

def dict2xml(xml_as_dict):
  return xmltodict.unparse(
    xml_as_dict,
    pretty=True,
    short_empty_elements=True,
    indent="  "
  )
  
def model2xml(model):
  return dict2xml(model.as_dict(with_tag=True))

def xml2model(xml):
  return Definitions.from_dict(xmltodict.parse(xml))

def sanitize_xml(xml):
  return dict2xml(xmltodict.parse(xml))

def uncamel(string):
  """
    (
      (?<=[a-z0-9])     preceeded by a lowercase or number
      [A-Z]             => capital
    |
      (?!^)             not at beginning
      [A-Z]             => capital
      (?=[a-z])         followed by a lower case letter
    )
    credit: https://stackoverflow.com/questions/1175208/#12867228
  """
  string = re.sub("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))", r" \1", string)
  return re.sub(" \\s*", " ", string) # remove multiple spacing

def slugify(string):
  return re.sub("-+", "-",
           re.sub(" +", "-",
             re.sub("[^a-z0-9- ]", "",
               re.sub("_", " ", uncamel(string)).lower()).strip()))



"""
 
The conditional builder allows for defining a process as a sequence of tasks,
each with a number of conditions that should be met before executing the task
 
The builder can construct a process flow, using the process builder, that
represents a flow diagram for that sequence of tasks with branching gates for
all conditions.
 
The builder can be "configured" using a Strategy defining how e.g. branches are
handled.
 
"""
 
import logging
 
import sys
 
from dataclasses import dataclass, field
from typing import List, Type, Union, Callable
 
from bpmn_tools.builder import process
from bpmn_tools import flow
 
logger = logging.getLogger(__name__)
 
ConditionKind = process.BranchKind
 
@dataclass
class Strategy():
  collapse_branches : bool = True
  default_branch_on : List[ConditionKind] = field(default_factory=lambda:[ConditionKind.XOR])
 
  def __getattr__(self, name):
    """
    wrap access to other classes/build-factory-function-classes, injecting
    this strategy if possible
    """
    def wrap(cls):
      def wrapper(*args, **kwargs):
        try:
          return cls(*args, **kwargs, strategy=self)
        except TypeError:
          return cls(*args, **kwargs)
      return wrapper
    return wrap(getattr(sys.modules[__name__], name))
 
@dataclass
class Condition():
  name    : str
  kind    : ConditionKind = ConditionKind.XOR
 
  def __eq__(self, other):
    return self.name == other.name and self.kind == other.kind
 
  def __str__(self):
    return f"{self.name}({self.kind.name})"
 
@dataclass
class ConditionSet():
  conditions : List[Condition] = field(default_factory=list)
  values     : List[List[str]] = field(default_factory=list)
 
  def __post_init__(self):
    for values in self.values:
      if len(values) != len(self.conditions):
        raise ValueError("each set of values should a value for each condition")
 
  def prune(self):
    # if all first values of a condition are None, it can be pruned 
    needs_pruning = True
    while needs_pruning:
      if not self.values:
        needs_pruning = False
      for values in self.values:
        if not values or values[0] is not None:
          needs_pruning = False
      if needs_pruning:
        self.conditions = self.conditions[1:]
        for index, values in enumerate(self.values):
          self.values[index] = values[1:]
 
  def __len__(self):
    return len(self.conditions)
 
  def __bool__(self):
    return len(self.conditions) > 0
 
  def __getitem__(self, index):
    return self.conditions[0]
 
  @property
  def first_non_none_values(self):
    return [
      values[0] for values in self.values if values[0] is not None
    ]
 
  def with_value(self, value):
    """
    return a clone with only value-sets of which the first value matches "value"
    """
    return ConditionSet(
      self.conditions.copy(),
      [ values for values in self.values if values[0] == value or values[0] is None ]
    )
 
  def without_first(self):
    """
    return a clone without first condition and its values
    """
    return ConditionSet(
      self.conditions[1:],
      [ values[1:] for values in self.values ]
    )
 
@dataclass
class Item():
  name       : str
  conditions : ConditionSet = field(default_factory=ConditionSet)
  cls        : Type[flow.Task] = flow.Task
  boundary   : Type[flow.EventDefinition] = None
  id         : Union[str, Callable] = None
  strategy   : Strategy = field(default_factory=Strategy)
 
  def to_dict(self):
    return self.name
 
  def to_process(self):
    return process.Task(
      name=self.name,
      id=self.id,
      cls=self.cls,
      boundary=self.boundary
    )
 
  def __eq__(self, other):
    return self.name == other.name
 
  def copy(self):
    return self.strategy.Item(self.name, self.conditions, self.cls, self.boundary, self.id)
 
  def with_value(self, value):
    item = self.copy()
    item.conditions = self.conditions.with_value(value)
    return item
 
  def without_first_condition(self):
    item = self.copy()
    item.conditions = self.conditions.without_first()
    return item
 
@dataclass
class Sequence():
  items    : List[Item] = field(default_factory=list)
  strategy : Strategy = field(default_factory=Strategy)
   
  def to_dict(self):
    return [ item.to_dict() for item in self.items ]
 
  def to_process(self, **kwargs):
    return process.Process([ item.to_process() for item in self.items ], **kwargs)
 
  def __eq__(self, other):
    try:
      return len(self.items) == len(other.items) and \
             all([ left == right for left, right in zip(self.items,other.items) ])
    except AttributeError:
      return False
 
  def __len__(self):
    return len(self.items)
 
  def __getitem__(self, index):
    return self.items[index]
 
  def append(self, item):
    self.items.append(item)
 
  def expand(self, *items):
    """
    takes a list of items and returns a list of expanded items
    items without conditions are simply returned
    items with conditions are wrapped in branches
    """
    self.items      = []          # reset
    items           = list(items) # ensure list
    prev_first_item = None        # track progress (avoid endless loop ;-))
 
    # prune all items' conditions -> remove all None-conditions up to first
    # actual condition
    for item in items:
      item.conditions.prune()
 
    logger.debug(f"creating sequence from {[item.name for item in items]}")
    while len(items) and items[0] is not prev_first_item:
      prev_first_item = items[0]
 
      # simply add condition-less items without further expansion
      while len(items) and not len(items[0].conditions):
        logger.debug(f"adding {items[0].name} to sequence")
        self.append(items.pop(0))
     
      # we encountered an item with a condition
      # collect leading sub-list of items that have the same first condition
      # and create a BranchedItem from them
      if len(items):
        branched_item  = self.strategy.BranchedItem(items[0].conditions[0])
        branched_items = []
       
        while len(items) and items[0].conditions \
              and items[0].conditions[0].name == branched_item.name \
              and items[0].conditions.first_non_none_values:
          branched_items.append(items.pop(0))
        branched_item.expand(*branched_items)
        self.append(branched_item)
    return self
 
@dataclass
class Branch():
  value    : tuple
  sequence : Sequence = field(default_factory=list)
  strategy : Strategy = field(default_factory=Strategy)
 
  def to_dict(self):
    return { ','.join(map(str,self.value)) : self.sequence.to_dict() }
 
  def to_process(self):
    return process.If(','.join(map(str,self.value)), self.sequence.to_process())
 
  def __eq__(self, other):
    return self.value == other.value and self.sequence == other.sequence
 
  def expand(self, *items):
    # only accept items with our condition
    for item in items:
      for values in item.conditions.values:
        if values[0] not in self.value and values[0] is not None:
          raise ValueError(f"{self.value} != {values[0]}")
    # create a sequence for the items, without the first condition, which
    # already brought us here (condition reduction up to no conditions left)
    items = [ item.without_first_condition() for item in items ]
    self.sequence = self.strategy.Sequence().expand(*items)
 
@dataclass
class BranchedItem():
  condition : Condition
  branches  : List[Branch] = field(default_factory=list)
  strategy  : Strategy = field(default_factory=Strategy)
 
  def __post_init__(self):
    print(self)
 
  @property
  def name(self):
    return self.condition.name
 
  def to_dict(self):
    return {
      str(self.condition) : [ branch.to_dict() for branch in self.branches ]
    }
 
  def to_process(self):
    return process.Branch(
      [ branch.to_process() for branch in self.branches ],
      label=self.condition.name,
      kind=self.condition.kind,
      default=self.condition.kind in self.strategy.default_branch_on
    )
 
  def __eq__(self, other):
    return self.condition == other.condition and \
           all([ left == right for left, right in zip(self.branches, other.branches) ])
 
  def __len__(self):
    return len(self.branches)
 
  def __getitem__(self, index):
    return self.branches[index]
 
  def expand(self, *items):
    """
    create branch for each condition-value
    expand matching items into branch
    """
    # only accept items with our condition
    for item in items:
      if item.conditions[0].name != self.name:
        raise ValueError(f"{self} only accepts items with condition {self.name}")
 
    logger.debug(f"creating branched item for {self.condition.name} from {[item.name for item in items]}")
    # create a set of the possible values for conditions
    values = sorted(set([
      values[0] for item in items for values in item.conditions.values
      if values[0] is not None
    ]))
    logger.debug(f"  with possible branch values: {values}")
 
    # create branches for each possible value
    for value in values:
      branch = self.strategy.Branch((value,))
      branch_items = []
      for item in items:
        branch_item = item.with_value(value)
        if branch_item.conditions.values:
          branch_items.append(branch_item)
      logger.debug(f"    - {value} : {[ item.name for item in branch_items ]}")
      branch.expand(*branch_items)
      self.branches.append(branch)
 
    # collapse branches that are the same (except for their value)
    if self.strategy.collapse_branches:
      collapsed_branches = []
      for index, branch in enumerate(self.branches):
        if index not in collapsed_branches:
          for other_index, other_branch in enumerate(self.branches[index+1:]):
            if branch != other_branch:
              if branch.sequence == other_branch.sequence:
                collapsed_branches.append(index+other_index+1)
                branch.value = branch.value + other_branch.value
      self.branches = [
        branch for index, branch in enumerate(self.branches)
        if index not in collapsed_branches
      ]



import io
from pathlib import Path
import xmltodict
import json

from bpmn_tools.tool import CLI

def test_loading_of_bpmn_from_stdin(monkeypatch):
  cli = CLI()
  folder = Path(__file__).resolve().parent / "models" 
  filepath = folder / "hello.bpmn"
  expected = filepath.read_text()
  monkeypatch.setattr("sys.stdin", io.StringIO(expected))
  cli.load()
  result_str = str(cli)
  result_export = cli.export()
  assert result_str == result_export
  assert xmltodict.parse(result_str) == xmltodict.parse(expected)

def test_loading_of_bpmn_from_load(monkeypatch):
  monkeypatch.setattr("sys.stdin", io.StringIO())
  cli = CLI()
  folder = Path(__file__).resolve().parent / "models"
  filepath = folder / "hello.bpmn"
  cli.load(filepath)
  result_str = str(cli)
  result_export = cli.export()
  assert result_str == result_export
  expected = filepath.read_text()
  assert xmltodict.parse(result_str) == xmltodict.parse(expected)

def test_loading_of_json_from_stdin(monkeypatch):
  cli = CLI()
  folder = Path(__file__).resolve().parent / "models"
  filepath = folder / "hello.json"
  json_str = filepath.read_text()
  monkeypatch.setattr("sys.stdin", io.StringIO(json_str))
  result = str(cli.load("-:json"))
  expected = (folder / "hello.bpmn").read_text()
  assert xmltodict.parse(result) == xmltodict.parse(expected)

def test_loading_of_json_from_load(monkeypatch):
  monkeypatch.setattr("sys.stdin", io.StringIO())
  cli = CLI()
  folder = Path(__file__).resolve().parent / "models"
  filepath = (folder / "hello.json")
  cli.load(filepath)
  result_str = str(cli)
  result_export = cli.export()
  assert result_str == result_export
  expected = (folder / "hello.bpmn").read_text()
  assert xmltodict.parse(result_str) == xmltodict.parse(expected)

def test_exporting_to_json(monkeypatch):
  monkeypatch.setattr("sys.stdin", io.StringIO())
  cli = CLI()
  folder = Path(__file__).resolve().parent / "models"
  filepath = folder / "hello.bpmn"
  result = str(cli.load(filepath).export("json"))
  expected = filepath.read_text()
  assert json.loads(result) == xmltodict.parse(expected)



import logging

import math

from bpmn_tools.flow import Gateway, Flow

from bpmn_tools import xml

logger = logging.getLogger(__name__)

class WayPoint(xml.Element):
  __tag__ = "di:waypoint"
  
  def __init__(self, x=0, y=0):
    super().__init__()
    self["x"] = str(int(x))
    self["y"] = str(int(y))

  @property
  def x(self):
    return int(self["x"])

  @property
  def y(self):
    return int(self["y"])

  @property
  def xy(self):
    return (self.x, self.y)

class Routable():
  def __init__(self, element=None):
    self.element = element

  @property
  def left(self):
    return self.element.x

  @property
  def center(self):
    return self.element.x + int(self.element.width/2)

  @property
  def right(self):
    return self.element.x + self.element.width

  @property
  def top(self):
    return self.element.y

  @property
  def middle(self):
    return self.element.y + int(self.element.height/2)

  @property
  def bottom(self):
    return self.element.y + self.element.height
  
  @property
  def north(self):
    return WayPoint(x=self.center, y=self.top)

  @property
  def east(self):
    return WayPoint(x=self.right, y=self.middle)

  @property
  def south(self):
    return WayPoint(x=self.center, y=self.bottom)

  @property
  def west(self):
    return WayPoint(x=self.left, y=self.middle)

  @property
  def gates(self):
    return [ self.north, self.east, self.south, self.west ]

class EdgeRoutingStrategy():
  """
  Base class for Edge Routing Strategies. Holds several routing supporting
  functions.
  """
  def __init__(self, flow=None):
    self.source = None
    self.target = None
    self.flow   = None
    self.given(flow)
    
  def given(self, new_flow=None):
    self.flow = new_flow
    if self.flow:
      self.source = Routable(self.flow.source)
      self.target = Routable(self.flow.target)
    return self

class Direct(EdgeRoutingStrategy):
  def route(self, flow):
    """
    slightly brute-force approach, computing distance between all src/trg gates
    selecting the shortest
    """
    self.given(flow)
    _, src, trg = min([
      ( math.dist(src.xy, trg.xy), src, trg )
      for src in self.source.gates
      for trg in self.target.gates
    ], key=lambda p: p[0])
    return [ src, trg ]

class Default(EdgeRoutingStrategy):
  """
  Default Edge Routing distinguishes between activities and gateways.
  Activities routes from side to side, while gateways are routed bottom to top
  
  TODO: reuse EdgeRoutingStrategy Routables
  """
  def route(self, flow):
    if isinstance(flow.source, Flow):
      logger.warning(f"routing is not supported for source Flow: {flow}")
      return []
    if isinstance(flow.target, Flow):
      logger.warning(f"routing is not supported for targer Flow: {flow}")
      return []
    self.given(flow)
    if isinstance(self.flow.source, Gateway) or isinstance(self.flow.target, Gateway):
      return self._route_gateway()
    return self._route_default()
  
  def _route_default(self):
    """
      Default routing routes directly from the right/middel exit of the source
      to the left/middle of the target.
    
      Unless the target lies completely above or below the source, then routing
      starts from the bottom/top to the top/bottom.

                                    T (above)
      0,0 --->                      ^
       |    X       S --> T     +---+
       | Y                      |
       v                        S
    """
    # determine position of target vs source
    below = self.flow.target.y > (self.flow.source.y + self.flow.source.height)
    above = (self.flow.target.y + self.flow.target.height) < self.flow.source.y

    if not (above or below): 
      children = [
        WayPoint(
          x=self.flow.source.x + self.flow.source.width,
          y=self.flow.source.y + int(self.flow.source.height/2)
        ),
        WayPoint(
          x=self.flow.target.x,
          y=self.flow.target.y + int(self.flow.target.height/2)
        )
      ]
    else:
      if below:
        top     = self.flow.source
        bottom  = self.flow.target
        reverse = False
      else:
        top     = self.flow.target
        bottom  = self.flow.source
        reverse = True
      children = [
        WayPoint(                               # bottom middle exit 
          x=top.x + int(top.width/2),
          y=top.y + top.height
        ),
        WayPoint(
          x=top.x + int(top.width/2),           # straight down to 17px
          y=bottom.y - 17
        ),
        WayPoint(                               # left/right to bottom
          x=bottom.x + int(bottom.width/2),
          y=bottom.y - 17
        ),
        WayPoint(
          x=bottom.x + int(bottom.width/2),     # top middle entry
          y=bottom.y
        )
      ]
      if reverse:
        children.reverse()
    return children
      
  def _route_gateway(self):
    """
      Gateway routing routes directly from the right/middle exit of the source
      to the left/middle of the target.
    
      Unless the target's middle lies above or below the gateway, then routing
      starts from the bottom/top to the left/middle.

                                    
      0,0 --->                      
       |    X       G --> T     +---> T
       | Y                      |
       v                        G
    """
    # determine position of target vs source
    below = self.flow.target.y > (self.flow.source.y + int(self.flow.source.height/2))
    above = (self.flow.target.y + int(self.flow.target.height/2)) < self.flow.source.y
    if not (above or below): 
      children = [
        WayPoint(
          x=self.flow.source.x + self.flow.source.width,
          y=self.flow.source.y + int(self.flow.source.height/2)
        ),
        WayPoint(
          x=self.flow.target.x,
          y=self.flow.target.y + int(self.flow.target.height/2)
        )
      ]
    else:
      if below:
        top     = self.flow.source
        bottom  = self.flow.target
        if isinstance(top, Gateway): # starting from GW?
          children = [
            WayPoint(                               # bottom middle exit 
              x=top.x + int(top.width/2),
              y=top.y + top.height
            ),
            WayPoint(
              x=top.x + int(top.width/2),           # straight down middle
              y=bottom.y + int(bottom.height/2)
            ),
            WayPoint(                               # left/right to left
              x=bottom.x,
              y=bottom.y + int(bottom.height/2)
            )
          ]
        else:
          children = [
            WayPoint(                               # middle right exit 
              x=top.x + top.width,
              y=top.y + int(top.height/2)
            ),
            WayPoint(
              x=bottom.x + int(bottom.width/2),     # right to middle
              y=top.y + int(top.height/2)
            ),
            WayPoint(                               # down to top
              x=bottom.x + int(bottom.width/2),
              y=bottom.y
            )
          ]
      else: # target is above
        top     = self.flow.target
        bottom  = self.flow.source
        # gateways route from top/bottom to side of tasks, but
        # from side to top/bottom of other gws
        if isinstance(bottom, Gateway) and not isinstance(top, Gateway):
          children = [
            WayPoint(                               # top middle exit 
              x=bottom.x + int(bottom.width/2),
              y=bottom.y
            ),
            WayPoint(
              x=bottom.x + int(bottom.width/2),     # straight up to middle
              y=top.y + int(top.height/2)
            ),
            WayPoint(                               # left/right to left
              x=top.x,
              y=top.y + int(top.height/2)
            )
          ]
        else:
          children = [
            WayPoint(                               # middle right exit 
              x=bottom.x + bottom.width,
              y=bottom.y + int(bottom.height/2)
            ),
            WayPoint(                               # right to middle
              x=top.x + int(top.width/2),
              y=bottom.y + int(bottom.height/2)
            ),
            WayPoint(                               # up to bottom
              x=top.x + int(top.width/2),
              y=top.y + top.height
            )
          ]
    return children


import logging
import sys

import xmltodict
from pathlib import Path
import json

from bpmn_tools import __version__
from bpmn_tools.notation import Definitions

from bpmn_tools.layout import * #noqa

logger = logging.getLogger(__name__)

class CLI():
  """
  thé command line interface to bpmn-tools: bpmn-tool
  
  enables reading/writing BPMN XML (or JSON) files and more (to come ;-))
  
    % bpmn-tool load tests/models/hello.bpmn 
    % bpmn-tool load - < tests/models/hello.bpmn
    % bpmn-tool load tests/models/hello.bpmn export json
  """
  def __init__(self):
    self._model = Definitions(id="definitions")

  def version(self):
    """
    provides bpmn-tools's version.
    """
    return __version__ # pragma: no cover
    
  def load(self, filepath="-"):
    """
    accepts a filepath to a file in one of the supported formats 
    provide "-" or "-:<format>" to load from stdin
    """
    if str(filepath)[0] == "-":
      src = sys.stdin.read()
      _, format = f"{filepath}:bpmn".split(":", 2)[0:2]
    else:  
      filepath = Path(filepath)
      format   = filepath.suffix[1:]
      src      = filepath.read_text()
    loader = getattr(self, f"_import_{format}")
    self._model = Definitions.from_dict(loader(src))
    return self

  def save(self, filepath):
    """
    accepts a filepath to a file and saves the current model to it, exported
    using the format provided by the filepath extension
    """
    filepath = Path(filepath)
    format   = filepath.suffix[1:]
    filepath.write_text(self.export(format))
    return self

  def export(self, format="bpmn"):
    """
    export the model (to stdout) in the requested format, default is BPMN
    """
    return getattr(self, f"_export_{format}")()

  def __str__(self):
    """
    this is triggered at the end if no explicit action (e.g. export is given)
    """
    return self.export()

  def layout(self, engine):
    """
    apply the requested layout engine
    """
    try:
      getattr(sys.modules[f"bpmn_tools.layout.{engine}"], "layout")(self._model)
    except KeyError:
      engines = ", ".join([
        name[len("bpmn_tools.layout."):] 
        for name in sys.modules.keys() if name.startswith("bpmn_tools.layout.")
      ])
      logger.error(f"unknown layout engine: `{engine}`. options are: {engines}")
      return ""
    return self

  # importer and exporters

  def _import_bpmn(self, bpmn_str):
    return xmltodict.parse(bpmn_str)

  def _import_json(self, json_str):
    return json.loads(json_str)

  def _export_bpmn(self):
    return xmltodict.unparse(self._model.as_dict(with_tag=True), pretty=True, indent="  ")

  def _export_json(self):
    return json.dumps(self._model.as_dict(with_tag=True), indent=2)



"""
  A simple lay-out-er.
  Supports processes looking like: 
    start -> task (-> task)* -> end
"""

import logging

import json

from bpmn_tools.flow          import Process, Start, End
from bpmn_tools.flow          import IntermediateCatch, IntermediateThrow
from bpmn_tools.flow          import Tasks, Gateways
from bpmn_tools.collaboration import Participant
from bpmn_tools.visitor       import Visitor, visiting

logger = logging.getLogger(__name__)

class LayoutVisitor(Visitor):
  def __init__(self):
    super().__init__()
    self.processes = {}
    self.process_participant = {}
    self._current_process = None
  
  def analyze(self, model):
    model.accept(self)
    logger.debug(json.dumps(self.processes, indent=2, default=str))
    return self
    
  @visiting(Process)
  def visit(self, process): # noqa
    logger.info(f"detecting elements in process: {process}")
    self.current_process = process

  @visiting(Participant) 
  def visit(self, participant): # noqa
    logger.info(f"found participant: {participant}")
    self.process_participant[participant.process.id] = participant
  
  @visiting(Start)
  def visit(self, event): # noqa
    self.current_process["start"] = event
    self._analyse_element(event)

  @visiting(IntermediateThrow, IntermediateCatch)
  def visit(self, event): # noqa
    self._analyse_element(event)

  @visiting(End)
  def visit(self, event): # noqa
    self.current_process["end"] = event
    self._analyse_element (event)

  @visiting(*Tasks)
  def visit(self, task): # noqa
    self._analyse_element(task)

  @visiting(*Gateways)
  def visit(self, gateway): # noqa
    self._analyse_element(gateway)

  @property
  def current_process(self):
    return self.processes[self._current_process.id]

  @current_process.setter
  def current_process(self, process):
    self._current_process = process
    if self._current_process.id not in self.processes:
      self.processes[self._current_process.id] = {
        "process"     : self._current_process,
        "height"      : 0,
        "elements"    : {},
        "start"       : None,
        "steps"       : {},
        "end"         : None
      }

  def _analyse_element(self, element):
    # keep index of all elements
    self.current_process["elements"][element.id] = element
    # record steps
    self.current_process["steps"][element.id] = [
      outgoing.target.id for outgoing in element.outgoing
    ]
    # track heighest element
    self.current_process["height"] = max([
      self.current_process["height"],
      element.height
    ])

  def layout(self):
    """
      start -> task (-> task)* -> end

      participant(x,y) = 160,80
      start(x,y) = (160+15+25,                     80+25+((80-36)/2)  (width,height=36)
      task(x,y)  = (160+15+25+36+50,               80+25)             (width=100, height=80)
      task(x,y)  = (160+15+25+36+50+100+50,        80+25)             (width=100, height=80)
      start(x,y) = (160+15+25+36+50+100+50+100+50, 80+25+((80-36)/2)  (width,height=36)
    """ # noqa
    START   = 160
    HEADER  = 30
    PADDING = 25
    SPACING = 30

    top = 80
    for process, analysis in self.processes.items():
      left = START
      self.process_participant[process].x = left
      self.process_participant[process].y = top

      if analysis["process"].laneset.lanes:
        lanes_height = 0
        for lane in analysis["process"].laneset.lanes:
          lane.x = left + HEADER
          lane.y = top + lanes_height
          lanes_height += lane.height
        self.process_participant[process].height = lanes_height
      else:
        pass

      left += HEADER + PADDING
      top += PADDING

      for step in self._order(analysis):
        step.x = left
        lane = analysis["process"].laneset.lane_of(step)
        if lane:
          step.y = PADDING + lane.y + (analysis["height"]-step.height) / 2
        else:
          step.y = top + (analysis["height"]-step.height) / 2
        left += step.width + SPACING

      width = left - START
      self.process_participant[process].width = width
      for lane in analysis["process"].laneset.lanes:
        lane.width  = width - HEADER
      top += self.process_participant[process].height

  def _order(self, analysis):
    """
    yield all elements in order:
      - start with any start element and recurse
      - loop over _all_ elements and recurse
      - only yielding each element once
    """
    handled = []
    def recurse(step):
      nonlocal handled
      if step in handled:
          return
      handled.append(step)
      yield step
      for next_step in analysis["steps"][step.id]:
        try:
          yield from recurse(analysis["elements"][next_step])
        except KeyError:
          pass

    # give priority to start-driven processes
    if analysis["start"]:
      yield from recurse(analysis["start"])

    # re-iterate all elements to cover loose ends
    for element in analysis["elements"].values():
      yield from recurse(element)

def layout(model):
  LayoutVisitor().analyze(model).layout()



from pathlib import Path

from bpmn_tools.builder.conditional import Item, Sequence, BranchedItem, Branch
from bpmn_tools.builder.conditional import ConditionSet, Condition, ConditionKind

from bpmn_tools import flow

folder = Path(__file__).resolve().parent / "models"

def test_single_item_without_conditions():
  item = Item("name")
  sequence = Sequence().expand(item)
  assert len(sequence) == 1
  assert sequence[0] == item

def test_multiple_items_without_conditions():
  items = [ Item("name 1"), Item("name 2") ]
  sequence = Sequence().expand(*items)
  assert len(sequence) == 2
  assert sequence[0] == items[0]
  assert sequence[1] == items[1]

def test_single_item_with_single_condition_and_single_value():
  item = Item(
    "name", ConditionSet( [ Condition("condition") ], [["value"]])
  )
  sequence = Sequence().expand(item)
  assert len(sequence) == 1
  assert isinstance(sequence[0], BranchedItem)
  branched_item = sequence[0]
  assert branched_item.name == "condition"
  assert len(branched_item) == 1
  branch = branched_item[0]
  assert branch.value == ("value",)
  assert len(branch.sequence) == 1
  assert branch.sequence[0].name == "name"

def test_single_item_with_single_condition_and_multiple_values_to_collapse():
  item = Item(
    "name", ConditionSet(
      [Condition("condition")], [["value 1"], ["value 2"]]
    )
  )
  sequence = Sequence().expand(item)
  assert len(sequence) == 1
  assert isinstance(sequence[0], BranchedItem)
  branched_item = sequence[0]
  assert branched_item.name == "condition"
  assert len(branched_item) == 1
  assert branched_item[0].value == ("value 1","value 2")
  assert branched_item[0].sequence[0].name == "name"

def test_single_item_with_multiple_conditions_and_multiple_value():
  item = Item(
    "name", ConditionSet(
      [Condition("condition 1"), Condition("condition 2")],
      [["value 1a", "value 1b"], ["value 2a", "value 2b"]]
    )
  )
  sequence = Sequence().expand(item)
  assert len(sequence) == 1
  assert isinstance(sequence[0], BranchedItem)
  branched_item = sequence[0]
  assert branched_item.name == "condition 1"
  assert len(branched_item) == 2
  assert isinstance(branched_item[0].sequence[0], BranchedItem)
  assert isinstance(branched_item[1].sequence[0], BranchedItem)

def test_a_more_complex_all_in_one_flow(compare, compare_model_to_file):
  sequence = Sequence().expand(
    Item("step 1"),
    Item("step 2", ConditionSet(
      [Condition("condition 1", ConditionKind.AND)],
      [["value 1"], ["value 2"]]
    )),
    Item("step 3", ConditionSet(
      [Condition("condition 1"), Condition("condition 2")],
      [["value 1", "value a"], ["value 3", "value b"], [None, "value c"]]
    )),
    Item("step 4"),
    Item("step 5", ConditionSet(
      [Condition("condition 1")],
      [["value 1"]]
    )),
    Item("step 6", ConditionSet(
      [Condition("condition 1")],
      [["value 1"]]
    )),    
  )
  compare(sequence.to_dict(), [
    "step 1",
    {
      "condition 1(AND)": [
        {
          "value 1": [
            "step 2",
            {
              "condition 2(XOR)": [
                { "value a,value c": [ "step 3" ] }
              ]
            }
          ]
        },
        {
          "value 2": [
            "step 2",
            {
              "condition 2(XOR)": [
                { "value c": [ "step 3" ] }
              ]
            }
          ]
        },
        {
          "value 3": [
            {
              "condition 2(XOR)": [
                { "value b,value c": [ "step 3" ] }
              ]
            }
          ]
        }
      ]
    },
    "step 4",
    {
      "condition 1(XOR)" : [
        { "value 1": [ "step 5", "step 6" ] }
      ]
    }
  ])

  process = sequence.to_process()
  model = process.render()

  filepath = folder / "conditional-builder.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_custom_bpmn_element_class_and_boundary(compare_model_to_file):
  sequence = Sequence().expand(
    Item("step 1", cls=flow.ManualTask, boundary=flow.TimerEventDefinition),
  )

  process = sequence.to_process()
  model = process.render()

  filepath = folder / "conditional-builder-with-custom-class-and-boundary.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_all_none_conditions_1(compare):
  sequence = Sequence().expand(
    Item("step 1"),
    Item("step 2", ConditionSet(
      [Condition("condition 1"), Condition("condition 2")],
      [
        ["value 1", "value a"],
        ["value 1", "value b"],
        ["value 2", None     ]
      ]  
    ))
  )
  compare(sequence.to_dict(), [
    "step 1",
    {
      "condition 1(XOR)": [
        {
          "value 1": [
            {
              "condition 2(XOR)": [
                {
                  "value a,value b": [
                    "step 2"
                  ]
                }
              ]
            }
          ]
        },
        {
          "value 2": [
            "step 2"
          ]
        }
      ]
    }
  ])

def test_all_none_conditions_2(compare):
  sequence = Sequence().expand(
    Item("step 1"),
    Item("step 2", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 1", "value a", None],
        ["value 1", "value b", None],
        ["value 2", None,      None]
      ]  
    ))
  )
  compare(sequence.to_dict(), [
    "step 1",
    {
      "condition 1(XOR)": [
        {
          "value 1": [
            {
              "condition 2(XOR)": [
                {
                  "value a,value b": [
                    "step 2"
                  ]
                }
              ]
            }
          ]
        },
        {
          "value 2": [
            "step 2"
          ]
        }
      ]
    }
  ])

def test_all_none_conditions(compare, compare_model_to_file):
  sequence = Sequence().expand(
    Item("step 1"),
    Item("step 2", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 1", "value a", None],
        ["value 1", "value b", None],
        ["value 2", None,      None]
      ]  
    )),
    Item("step 3", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 3", None, None],
        ["value 4", None, None]
      ]  
    )),
    Item("step 4", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 3", None, None],
        ["value 4", None, None]
      ]  
    )),
    Item("step 5", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 1", None, "something"],
        ["value 2", None, None]
      ]  
    )),
    Item("step 6", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        ["value 2", None, None]
      ]  
    )),
    Item("step 7", ConditionSet(
      [Condition("condition 1"), Condition("condition 2"), Condition("condition 3")],
      [
        [ None, None, "something"]
      ]  
    ))
  )
  compare(sequence.to_dict(), [
    "step 1",
    {
      "condition 1(XOR)": [
        {
          "value 1": [
            {
              "condition 2(XOR)": [
                {
                  "value a,value b": [
                    "step 2"
                  ]
                }
              ]
            },
            {
              "condition 3(XOR)": [
                {
                  "something": [
                    "step 5"
                  ]
                }
              ]
            }
          ]
        },
        {
          "value 2": [
            "step 2",
            "step 5",
            "step 6"
          ]
        },
        {
          "value 3,value 4": [
            "step 3",
            "step 4"
          ]
        }
      ]
    },
    {
      "condition 3(XOR)": [
        {
          "something": [
            "step 7"
          ]
        }
      ]
    }
  ])

  process = sequence.to_process()
  model = process.render()

  filepath = folder / "conditional-builder-with-various-nones.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_collapse_sequential_branched_items(compare, compare_model_to_file):
  sequence = Sequence().expand(
    Item("step 1"),
    Item("step 2", ConditionSet(
      [Condition("condition 1"), Condition("condition 2")],
      [
        [None, "value 1"]
      ]  
    )),
    Item("step 3", ConditionSet(
      [Condition("condition 1"), Condition("condition 2")],
      [
        [None, "value 2"],
        [None, "value 3"]
      ]  
    )),
    Item("step 4")
  )
  compare(sequence.to_dict(), [
    "step 1",
    {
      "condition 2(XOR)": [
        {
          "value 1": [ "step 2" ]
        },
        {
          "value 2,value 3": [ "step 3" ]
        }
      ]
    },
    "step 4"
  ])

  process = sequence.to_process()
  model = process.render()

  filepath = folder / "conditional-builder-collapse_sequential_branched_items.bpmn"
  compare_model_to_file(model, filepath, save_to=f"{filepath.stem}-test.bpmn")

def test_adding_process_properties(compare):
  sequence = Sequence().expand(Item("step 1"))
  process = sequence.to_process(name="with properties", starts=True, ends=True)
  assert len(process.children) == 1
  assert process.children[0].name == "step 1"
  assert process.name == "with properties"
  assert process.starts
  assert process.ends

def test_conditions_require_same_number_of_conditions_and_values():
  try:
    ConditionSet(["condition 1", "condition 2"], [[[1, 2, 3]]])
    assert False, "number of values must match number of conditions"
  except ValueError:
    pass

def test_branch_and_item_condition_value_should_match():
  item = Item("name", ConditionSet(["condition"], [[ "something" ]] ))
  try:
    Branch(("value",)).expand(item)
    assert False, "branch should only accept items with same condition value"
  except ValueError:
    pass

def test_branched_item_and_item_condition_should_match():
  item = Item("name", ConditionSet([Condition("condition")], [[ "something" ]] ))
  try:
    BranchedItem(Condition(["other condition"])).expand(item)
    assert False, "branched item should only accept items with same condition"
  except ValueError:
    pass
  
def test_provide_id_for_task():
  item = Item("name")                                 # task_0
  assert item.to_process().id == "task_0"

  item = Item("name", id="testing")                   # task_1
  assert item.to_process().id == "testing"
  
  item = Item("name", id=lambda d: f"prefixed_{d}")   # task_2
  assert item.to_process().id == "prefixed_task_2"

def test_copy():
  item = Item(
    "name", ConditionSet(), flow.ManualTask, flow.TimerEventDefinition, "id"
  )
  copy = item.copy()
  assert copy.name == item.name
  assert copy.conditions is item.conditions
  assert copy.cls == item.cls
  assert copy.boundary == item.boundary
  assert copy.id == item.id
  


"""

Process builder builds single-process, forward branching-flow-like structures:


                           | -> [ Task ] ------------------- |
  S -> [ Task ] -> < Branch >                               < > -> [ Task ] -> S
                           | -> < Branch > ----------- < > - | 
                                        | -> [ Task ] - |   
""" 

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Dict, Type, Union, Callable

from bpmn_tools import flow

from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.notation      import Definitions
from bpmn_tools.colors        import Red, Green, Orange, Blue

from bpmn_tools.util import slugify

PADDING               =  10
FLOW_SPACE            =  20
MODEL_OFFSET_X        = 150
MODEL_OFFSET_Y        =  80
DEFAULT_BRANCH_HEIGHT =  75
STANDARD_TASK_HEIGHT  =  80

def NoColor(bpmn):
  return bpmn

class Color(Enum):
  NONE   : NoColor
  RED    : Red
  GREEN  : Green
  ORANGE : Orange
  BLUE   : Blue

@dataclass
class Step():
  children: List["Step"] = field(default_factory=list)
  label   : str   = None
  color   : Color = NoColor
  
  @property
  def height(self):
    raise NotImplementedError # pragma: no cover

  @property
  def width(self):
    raise NotImplementedError # pragma: no cover

  def render(self, x=0, y=0):
    raise NotImplementedError # pragma: no cover

  def shape(self, cls, **kwargs):
    return self.color(cls(**kwargs))

@dataclass
class Task(Step):
  name    : str = ""
  cls     : Type[flow.Task] = flow.Task
  args    : Dict = field(default_factory=dict)
  boundary: Type[flow.EventDefinition] = None
  id      : Union[str, Callable] = None
  _id     : Union[str, Callable] = field(init=False, repr=False, default=None)
  _std_id : str = field(init=False, repr=False, default=None)

  tasks = 0 # class

  @classmethod
  def reset(cls):
    cls.tasks = 0

  @property
  def id(self):
    if self._id:
      if callable(self._id):
        return self._id(self._std_id)
      else:
        return self._id
    return self._std_id
  
  @id.setter
  def id(self, new_id):
    # courtesy: https://stackoverflow.com/a/61480946
    if type(new_id) is property:
      # initial value not specified, use default
      new_id = Task._id
    self._id = new_id

  def __post_init__(self):
    self.args["name"] = self.name
    self._std_id = f"task_{self.tasks}"
    self.__class__.tasks += 1
    self.element = self.shape(self.cls, id=self.id, **self.args)
  
  @property
  def height(self):
    return self.element.height + PADDING * 2

  @property
  def width(self):
    return self.element.width + PADDING * 2

  def render(self, x=0, y=0):
    self.element.x = x + PADDING
    self.element.y = y + PADDING
    shapes = [ self.element ]
    if self.boundary:
      shapes.append(flow.BoundaryEvent(
        id=f"boundary-event-{self.element.id}",
        definition=self.boundary(id=f"{slugify(self.boundary.__name__)}-{self.element.id}"),
        on=self.element
      ))
    return (shapes, [])

  @property
  def root(self):
    return self.element

  @property
  def tail(self):
    return self.element

def connect(source, target, label=None):
  if source == target:
    raise ValueError("connect: {source} == {target}")
  return flow.Flow(id=f"flow_{source.id}_{target.id}", source=source, target=target, name=label)

@dataclass
class Process(Step):
  name   : str  = ""
  starts : bool = False
  ends   : bool = False
  _root  = None
  _tail  = None

  def __post_init__(self):
    if self.starts:
      self._root = self.shape(flow.Start, id="start")
    if self.ends:
      self._tail = self.shape(flow.End, id="end")

  def add(self, step):
    self.children.append(step)
    return self

  def extend(self, steps):
    for step in steps:
      self.add(step)
    return self

  @property
  def height(self):
    return max([ step.height for step in self.children ])

  @property
  def width(self):
    total_width = sum([ step.width for step in self.children ])
    total_width += FLOW_SPACE * (len(self.children) - 1)
    if self.starts:
      total_width += self.root.width + FLOW_SPACE
    if self.ends:
      total_width += self.tail.width + FLOW_SPACE
    return total_width 

  def render(self, x=None, y=None):
    # if we're called without positioning, we're adding diagram boilerplate
    wrap = x is None or y is None
    if wrap:
      x = 50 + MODEL_OFFSET_X # header of participant
      y = 0  + MODEL_OFFSET_Y
    
    shapes = []
    flows  = []
    prev = None

    # add optional start event
    if self.starts:
      self.root.x = x
      x += self.root.width
      self.root.y = y + int((STANDARD_TASK_HEIGHT + PADDING * 2)/2) - (self.root.height/2)
      shapes.append(self.root)
      x += FLOW_SPACE

    # add steps
    for step in self.children:
      more_shapes, more_flows = step.render(x=x, y=y)
      shapes.extend(more_shapes)
      flows.extend(more_flows)
      x += step.width + FLOW_SPACE
      if prev:
        flows.append(connect(source=prev.tail, target=step.root))
      elif self.starts:
        flows.append(connect(source=self.root, target=step.root))
      prev = step

    # add optional end event
    if self.ends:
      shapes.append(self.tail)
      # attach to tail of last child
      last_shape = self.children[-1].tail
      self.tail.x = x
      self.tail.y = last_shape.y + (last_shape.height/2) - (self.tail.height/2)
      # add flow to end
      flows.append(connect(source=last_shape, target=self.tail))

    # optinally wrap it all
    if wrap:
      process = flow.Process(id="process").extend(shapes).extend(flows)
      participant = Participant(self.name, process, id="participant")
      # participant shapes the lane
      participant.x = MODEL_OFFSET_X
      participant.y = MODEL_OFFSET_Y
      participant.width  = self.width + 50 + PADDING
      participant.height = self.height
      collaboration = Collaboration(id="collaboration").append(participant)
      return Definitions(id="definitions").extend([
        process,
        collaboration,
        Diagram(
          id="diagram",
          plane=Plane(id="plane", element=collaboration)
        )
      ])
    return (shapes, flows)

  @property
  def root(self):
    if self._root:
      return self._root
    return self.children[0].root

  @property
  def tail(self):
    if self._tail:
      return self._tail
    return self.children[-1].tail

class BranchKind(Enum):
  XOR = flow.ExclusiveGateway
  OR  = flow.InclusiveGateway
  AND = flow.ParallelGateway

# utility wrapper for an even more fluid API ;-)
def If(condition, step):
  return (step, condition)

@dataclass
class Branch(Step):
  default  : str = None
  kind     : BranchKind = BranchKind.XOR

  gws = 0

  def __post_init__(self):
    global gws
    self.root = self.shape(self.kind.value, id=f"gateway_start_{self.gws}", name=self.label)
    self.tail = self.shape(self.kind.value, id=f"gateway_end_{self.gws}")
    self.__class__.gws += 1
    # ensure all children to ensure they are tuples
    self.children = list(map(lambda c: c if type(c) is tuple else (c,None), self.children))

  @classmethod
  def reset(cls):
    cls.gws = 0

  def add(self, branch, condition=None):
    self.children.append((branch, condition))
    return self

  @property
  def height(self):
    total_height = sum([ branch.height for branch, _ in self.children ])
    if self.default:
      total_height += DEFAULT_BRANCH_HEIGHT
    return total_height

  @property
  def width(self):
    total_width = max([ branch.width for branch, _ in self.children ])
    total_width += self.root.width + self.tail.width
    total_width += FLOW_SPACE * 2
    return total_width

  def render(self, x=0, y=0):
    shapes = []
    flows  = []
    self.root.x = x
    self.root.y = y + int((STANDARD_TASK_HEIGHT + PADDING * 2)/2) - (self.root.height/2)
    shapes.append(self.root)
    shapes.append(self.tail)
    x += FLOW_SPACE
    if self.default:
      flows.append(connect(source=self.root, target=self.tail))
      y += DEFAULT_BRANCH_HEIGHT
    self.tail.x = x + self.width - self.tail.width - FLOW_SPACE
    self.tail.y = self.root.y
    x += self.root.width
    for branch, condition in self.children:
      more_shapes, more_flows = branch.render(x=x, y=y)
      shapes.extend(more_shapes)
      flows.extend(more_flows)
      y += branch.height
      flows.append(connect(source=self.root,   target=branch.root, label=condition))
      flows.append(connect(source=branch.tail, target=self.tail))
    return shapes, flows



"""
  A lay-out-er that uses Graphviz as a layout engine to layout activities in a
  process. If multiple processes are part of the model, each is laid out.
"""
 
import logging
 
import json
 
import graphviz
 
from bpmn_tools.flow          import Process
from bpmn_tools.flow          import Tasks
from bpmn_tools.collaboration import Participant

from bpmn_tools.visitor       import Visitor, visiting

from bpmn_tools.layout        import routing
 
logger = logging.getLogger(__name__)
 
class LayoutVisitor(Visitor):
  def __init__(self, g=None):
    super().__init__()
    self.processes = {}
    self.process_participant = {}
    self._current_process = None
    self.g = g if g else graphviz.Graph("", engine="sfdp", graph_attr={"repulsiveforce" : "10"})
 
  def analyze(self, model):
    model.accept(self)
    logger.debug(json.dumps(self.processes, indent=2, default=str))
    return self
   
  @visiting(Process)
  def visit(self, process): # noqa
    logger.debug(f"detecting elements in process: {process}")
    self.current_process = process
 
  @visiting(Participant)
  def visit(self, participant): # noqa
    logger.debug(f"found participant: {participant}")
    self.process_participant[participant.process.id] = participant
 
  @visiting(*Tasks)
  def visit(self, task): # noqa
    self._analyse_element(task)
 
  @property
  def current_process(self):
    return self.processes[self._current_process.id]
 
  @current_process.setter
  def current_process(self, process):
    self._current_process = process
    if self._current_process.id not in self.processes:
      self.processes[self._current_process.id] = {
        "process"      : self._current_process,
        "elements"     : {},
        "interactions" : []
      }
 
  def _analyse_element(self, element):
    self.current_process["elements"][element.id] = element
    self.current_process["interactions"].extend([
      (element.id, outgoing.target.id) for outgoing in element.outgoing
    ])
 
  def layout(self):
    START   = 160
    HEADER  = 30
    PADDING = 25
    SPACING = 30
 
    top = 80
    for process_id, analysis in self.processes.items():
      # apply process positions
      self.process_participant[process_id].x = START
      self.process_participant[process_id].y = top
 
      # activity offsets
      left = START + HEADER + PADDING
      top += PADDING
 
      positions = self._make_gv(analysis)
      logger.debug(json.dumps(positions, indent=2, default=str))
     
      width  = 0
      height = 0
      for element_id, x, y in positions:
        width  = max([width, x + analysis["elements"][element_id].width])
        height = max([height, y + analysis["elements"][element_id].height])
        analysis["elements"][element_id].x = x + left
        analysis["elements"][element_id].y = y + top
 
      self.process_participant[process_id].width  = width + HEADER + PADDING * 2
      self.process_participant[process_id].height = height + PADDING * 2
     
      top += height + PADDING + SPACING
 
  def _make_gv(self, process):
    for element in process["elements"].keys():
      self.g.node(element)
 
    for source, target in process["interactions"]:
      self.g.edge(source, target)
 
    self.g.render()
    gv = json.loads(self.g.pipe("json").decode())
 
    positions = [
      (obj["name"],) + tuple([int(float(pos)) for pos in obj["pos"].split(",")])
      for obj in gv["objects"]
    ]
   
    # normalize
    min_x = min([position[1] for position in positions])
    min_y = min([position[2] for position in positions])
 
    return [
      (position[0], position[1] - min_x, position[2] - min_y)
      for position in positions
    ]

def layout(model, g=None):
  LayoutVisitor(g=g).analyze(model).layout()
  model.apply(routing.Direct())



from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Tasks
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.layout        import simple

def test_all_task_types(compare_model_to_file):
  process = Process(id="process").extend(
    [ cls(cls.__name__, id=cls.__name__) for cls in Tasks ]
  )

  collaboration = Collaboration(id="collaboration").append(
    Participant("all tasks", process, id="participant")
  )

  model = Definitions(id="definitions").extend([
    process,
    collaboration,
  ])

  model.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  simple.layout(model)

  compare_model_to_file(model, "layout-all-tasks.bpmn", save_to="layout-all-tasks-test.bpmn")



from bpmn_tools.flow          import Process, Flow, Lane
from bpmn_tools.flow          import Task
from bpmn_tools.flow          import ExclusiveGateway
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.notation      import Definitions
from bpmn_tools.diagrams      import Diagram, Plane
from bpmn_tools.layout        import simple

def test_issue_1(compare_model_to_file):
  # minimal example to illustrate raise problem, based on submitted example
  # https://github.com/christophevg/bpmn-tools/issues/1

  a = Task("task a", id="a")
  b = Task("task b", id="b")

  decision = ExclusiveGateway("gateway", id="gateway")

  lane1 = Lane("lane", id="lane")
  lane1.extend([a, decision, b])

  flow1 = Flow(source=a,        target=decision, id="flow1")
  flow2 = Flow(source=decision, target=b,        id="flow2")

  process = Process(id="process").extend([ lane1, flow1, flow2 ])

  collaboration = Collaboration(id="Colaboration").extend([
    Participant("participant", process, id="participant")
  ])

  definitions = Definitions(id="definitions").extend([
    process,
    collaboration
  ])

  definitions.append(
    Diagram( id="diagram", plane=Plane(id="plane", element=collaboration) )
  )

  simple.layout(definitions)

  compare_model_to_file(definitions, "issues/1.bpmn", save_to="issues/1-test.bpmn")



from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Task, Flow, MessageFlow
from bpmn_tools.diagrams      import Diagram, Plane

from bpmn_tools.layout        import simple

def test_recursive_simple_layout(compare_model_to_file):
  
  activities = [
    Task("step 1", id="step1"),
    Task("step 2", id="step2"),
    Task("step 3", id="step3"),
    Task("step 4", id="step4")
  ]

  process1 = Process(id="process1").extend(activities).extend([
    Flow(id="flow_step1_step2", source=activities[0], target=activities[1]),
    Flow(id="flow_step1_step3", source=activities[0], target=activities[2]),
    Flow(id="flow_step1_step4", source=activities[0], target=activities[3])
  ])

  remote_activity = Task("step 5", id="step5")
  process2 = Process(id="process2").append(remote_activity)
  
  collaboration = Collaboration(id="collaboration").extend([
    Participant("process 1", process1, id="participant_process1"),
    Participant("process 2", process2, id="participant_process2")
  ]).append(
    MessageFlow(id="flow_step1_step5", source=activities[0], target=remote_activity)
  )

  definitions = Definitions(id="definitions").extend([
    process1,
    process2,
    collaboration
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )
  
  simple.layout(definitions)

  compare_model_to_file(
    definitions,
    "recursive_simple_layout.bpmn",
    save_to="recursive_simple_layout-test.bpmn"
  )
